import inspect
from dataclasses import dataclass
from functools import cached_property
from itertools import chain
from pathlib import Path
from typing import Callable, Dict, Mapping, Union

import h5py
import numpy as np
import pandas as pd
from anndata import AnnData
from anndata._core.index import Index, _normalize_indices
from anndata._core.sparse_dataset import SparseDataset
from anndata._core.views import _resolve_idx
from anndata._io.h5ad import read_dataframe_legacy as read_dataframe_legacy_h5
from anndata._io.specs.registry import get_spec, read_elem, read_elem_partial
from anndata.compat import _read_attr
from fsspec.core import OpenFile
from lamindb_setup.dev.upath import UPath, infer_filesystem
from lnschema_core import File

from lamindb.dev.storage.file import filepath_from_file


# try csr for groups with no encoding_type
class SparseDatasetCSR(SparseDataset):
    @property
    def format_str(self) -> str:
        return "csr"


# zarr and SparseDatasetCSR have problems with full selection
def _subset_sparse(sparse_ds: Union[SparseDatasetCSR, SparseDataset], indices):
    has_arrays = isinstance(indices[0], np.ndarray) or isinstance(
        indices[1], np.ndarray
    )
    if not has_arrays and indices == (slice(None), slice(None)):
        return sparse_ds.to_memory()
    else:
        return sparse_ds[indices]


def get_module_name(obj):
    return inspect.getmodule(obj).__name__.partition(".")[0]


def _records_to_df(obj):
    if isinstance(obj, pd.DataFrame):
        return obj

    if hasattr(obj, "dtype") and obj.dtype.names is not None:
        formats = []
        for name, (dt, _) in obj.dtype.fields.items():
            if dt.char == "S":
                new_dt = str(dt).replace("S", "U")
            else:
                new_dt = dt
            formats.append((name, new_dt))
        df = pd.DataFrame(obj.astype(formats, copy=False))
        for index_name in ("index", "_index"):
            if index_name in df.columns:
                return df.set_index(index_name)
            return df
    else:
        return obj


class Registry:
    def __init__(self):
        self._registry = {}
        self._openers = {}

    def register_open(self, module: str):
        def wrapper(func: Callable):
            self._openers[module] = func
            return func

        return wrapper

    def open(self, module: str, *args, **kwargs):
        if module in self._openers:
            return self._openers[module](*args, **kwargs)
        else:
            raise ValueError(f"Module {module} not found, please install it.")

    def register(self, module: str):
        def wrapper(func: Callable):
            func_name = func.__name__
            if func_name not in self._registry:
                self._registry[func_name] = {}
            self._registry[func_name][module] = func
            return func

        return wrapper

    def __getattr__(self, func_name: str):
        def wrapper(*args, **kwargs):
            func_registry = self._registry[func_name]
            for arg in chain(args, kwargs.values()):
                arg_module = get_module_name(arg)
                if arg_module in func_registry:
                    return func_registry[arg_module](*args, **kwargs)
            raise ValueError(f"{func_name} is not registered for this module.")

        return wrapper


# storage specific functions should be registered and called through the registry
registry = Registry()


@registry.register_open("h5py")
def open(filepath: Union[UPath, Path, str]):
    fs, file_path_str = infer_filesystem(filepath)
    conn = fs.open(file_path_str, mode="rb")
    try:
        storage = h5py.File(conn, mode="r")
    except Exception as e:
        conn.close()
        raise e
    return conn, storage


@registry.register("h5py")
def read_dataframe(elem: Union[h5py.Dataset, h5py.Group]):
    if isinstance(elem, h5py.Dataset):
        return read_dataframe_legacy_h5(elem)
    else:
        return read_elem(elem)


@registry.register("h5py")
def safer_read_partial(elem, indices):
    if get_spec(elem).encoding_type == "":
        if isinstance(elem, h5py.Dataset):
            dims = len(elem.shape)
            if dims == 2:
                return elem[indices]
            elif dims == 1:
                if indices[0] == slice(None):
                    return elem[indices[1]]
                elif indices[1] == slice(None):
                    return elem[indices[0]]
        elif isinstance(elem, h5py.Group):
            try:
                ds = SparseDatasetCSR(elem)
                return _subset_sparse(ds, indices)
            except Exception:
                pass
        raise ValueError(
            "Can not get a subset of the element of type"
            f" {type(elem).__name__} with an empty spec."
        )
    else:
        return read_elem_partial(elem, indices=indices)


@registry.register("h5py")
def keys(storage: h5py.File):
    attrs_keys: Dict[str, list] = {}
    for attr in storage.keys():
        if attr == "X":
            continue
        attr_obj = storage[attr]
        if attr in ("obs", "var") and isinstance(attr_obj, h5py.Dataset):
            keys = list(attr_obj.dtype.fields.keys())
        else:
            keys = list(attr_obj.keys())
        if len(keys) > 0:
            attrs_keys[attr] = keys
    return attrs_keys


ArrayTypes = [h5py.Dataset]
GroupTypes = [h5py.Group]
StorageTypes = [h5py.File]


ZARR_INSTALLED = False
try:
    import zarr

    ZARR_INSTALLED = True
except ImportError:
    pass

if ZARR_INSTALLED:
    from anndata._io.zarr import read_dataframe_legacy as read_dataframe_legacy_zarr

    ArrayTypes.append(zarr.Array)
    GroupTypes.append(zarr.Group)
    StorageTypes.append(zarr.Group)

    @registry.register_open("zarr")
    def open(filepath: Union[UPath, Path, str]):  # noqa
        fs, file_path_str = infer_filesystem(filepath)
        conn = None
        storage = zarr.open(fs.get_mapper(file_path_str, check=True), mode="r")
        return conn, storage

    @registry.register("zarr")
    def read_dataframe(elem: Union[zarr.Array, zarr.Group]):  # noqa
        if isinstance(elem, zarr.Array):
            return read_dataframe_legacy_zarr(elem)
        else:
            return read_elem(elem)

    @registry.register("zarr")
    def safer_read_partial(elem, indices):  # noqa
        encoding_type = get_spec(elem).encoding_type
        if encoding_type == "":
            if isinstance(elem, zarr.Array):
                dims = len(elem.shape)
                if dims == 2:
                    return elem.oindex[indices]
                elif dims == 1:
                    if indices[0] == slice(None):
                        return elem.oindex[indices[1]]
                    elif indices[1] == slice(None):
                        return elem.oindex[indices[0]]
            elif isinstance(elem, zarr.Group):
                try:
                    ds = SparseDatasetCSR(elem)
                    return _subset_sparse(ds, indices)
                except Exception:
                    pass
            raise ValueError(
                "Can not get a subset of the element of type"
                f" {type(elem).__name__} with an empty spec."
            )
        else:
            if encoding_type in ("csr_matrix", "csc_matrix"):
                ds = SparseDataset(elem)
                return _subset_sparse(ds, indices)
            else:
                return read_elem_partial(elem, indices=indices)

    # this is needed because accessing zarr.Group.keys() directly is very slow
    @registry.register("zarr")
    def keys(storage: zarr.Group):  # noqa
        paths = storage._store.keys()

        attrs_keys: Dict[str, list] = {}
        obs_var_arrays = []

        for path in paths:
            if path in (".zattrs", ".zgroup"):
                continue
            parts = path.split("/")
            if len(parts) < 2:
                continue
            attr = parts[0]
            key = parts[1]

            if attr == "X":
                continue

            if attr in ("obs", "var"):
                if attr in obs_var_arrays:
                    continue
                if key == ".zarray":
                    attrs_keys.pop(attr, None)
                    obs_var_arrays.append(attr)

            if attr not in attrs_keys:
                attrs_keys[attr] = []

            if key in (".zattrs", ".zgroup", ".zarray"):
                continue
            attr_keys = attrs_keys[attr]
            if key not in attr_keys:
                attr_keys.append(key)

        for attr in obs_var_arrays:
            attrs_keys[attr] = list(storage[attr].dtype.fields.keys())

        return {attr: keys for attr, keys in attrs_keys.items() if len(keys) > 0}


ArrayTypes = tuple(ArrayTypes)  # type: ignore
GroupTypes = tuple(GroupTypes)  # type: ignore
StorageTypes = tuple(StorageTypes)  # type: ignore


ArrayType = Union[ArrayTypes]  # type: ignore
GroupType = Union[GroupTypes]  # type: ignore
StorageType = Union[StorageTypes]  # type: ignore


def _to_memory(elem):
    if isinstance(elem, ArrayTypes):
        return elem[()]
    elif isinstance(elem, SparseDataset):
        return elem.to_memory()
    else:
        return elem


def _try_backed_full(elem):
    # think what to do for compatibility with old var and obs
    if isinstance(elem, ArrayTypes):
        return elem

    if isinstance(elem, GroupTypes):
        encoding_type = get_spec(elem).encoding_type
        if encoding_type in ("csr_matrix", "csc_matrix"):
            return SparseDataset(elem)
        if "h5sparse_format" in elem.attrs:
            return SparseDataset(elem)
        if encoding_type == "" and "indptr" in elem:
            return SparseDatasetCSR(elem)

    return read_elem(elem)


def _safer_read_index(elem):
    if isinstance(elem, GroupTypes):
        return pd.Index(read_elem(elem[_read_attr(elem.attrs, "_index")]))
    elif isinstance(elem, ArrayTypes):
        indices = None
        for index_name in ("index", "_index"):
            if index_name in elem.dtype.names:
                indices = elem[index_name]
                break
        if indices is not None and len(indices) > 0:
            if isinstance(indices[0], bytes):
                indices = np.frompyfunc(lambda x: x.decode("utf-8"), 1, 1)(indices)
            return pd.Index(indices)
        else:
            raise ValueError("Indices not found.")
    else:
        raise ValueError(f"Unknown elem type {type(elem)} when reading indices.")


class _MapAccessor:
    def __init__(self, elem, name, indices=None):
        self.elem = elem
        self.indices = indices
        self.name = name

    def __getitem__(self, key):
        if self.indices is None:
            return _try_backed_full(self.elem[key])
        else:
            return registry.safer_read_partial(self.elem[key], indices=self.indices)

    def keys(self):
        return list(self.elem.keys())

    def __repr__(self):
        """Description of the _MapAccessor object."""
        descr = f"Accessor for the AnnData attribute {self.name}"
        descr += f"\n  with keys: {self.keys()}"
        return descr


class _AnnDataAttrsMixin:
    storage: StorageType
    _attrs_keys: Mapping[str, list]

    @cached_property
    def obs(self) -> pd.DataFrame:
        if "obs" not in self._attrs_keys:
            return None
        indices = getattr(self, "indices", None)
        if indices is not None:
            indices = (indices[0], slice(None))
            obj = registry.safer_read_partial(self.storage["obs"], indices=indices)  # type: ignore # noqa
            return _records_to_df(obj)
        else:
            return registry.read_dataframe(self.storage["obs"])  # type: ignore

    @cached_property
    def var(self) -> pd.DataFrame:
        if "var" not in self._attrs_keys:
            return None
        indices = getattr(self, "indices", None)
        if indices is not None:
            indices = (indices[1], slice(None))
            obj = registry.safer_read_partial(self.storage["var"], indices=indices)  # type: ignore # noqa
            return _records_to_df(obj)
        else:
            return registry.read_dataframe(self.storage["var"])  # type: ignore

    @cached_property
    def uns(self):
        if "uns" not in self._attrs_keys:
            return None
        return read_elem(self.storage["uns"])

    @cached_property
    def X(self):
        indices = getattr(self, "indices", None)
        if indices is not None:
            return registry.safer_read_partial(self.storage["X"], indices=indices)
        else:
            return _try_backed_full(self.storage["X"])

    @cached_property
    def obsm(self):
        if "obsm" not in self._attrs_keys:
            return None
        indices = getattr(self, "indices", None)
        if indices is not None:
            indices = (indices[0], slice(None))
        return _MapAccessor(self.storage["obsm"], "obsm", indices)

    @cached_property
    def varm(self):
        if "varm" not in self._attrs_keys:
            return None
        indices = getattr(self, "indices", None)
        if indices is not None:
            indices = (indices[1], slice(None))
        return _MapAccessor(self.storage["varm"], "varm", indices)

    @cached_property
    def obsp(self):
        if "obsp" not in self._attrs_keys:
            return None
        indices = getattr(self, "indices", None)
        if indices is not None:
            indices = (indices[0], indices[0])
        return _MapAccessor(self.storage["obsp"], "obsp", indices)

    @cached_property
    def varp(self):
        if "varp" not in self._attrs_keys:
            return None
        indices = getattr(self, "indices", None)
        if indices is not None:
            indices = (indices[1], indices[1])
        return _MapAccessor(self.storage["varp"], "varp", indices)

    @cached_property
    def layers(self):
        if "layers" not in self._attrs_keys:
            return None
        indices = getattr(self, "indices", None)
        return _MapAccessor(self.storage["layers"], "layers", indices)

    @property
    def obs_names(self):
        return self._obs_names

    @property
    def var_names(self):
        return self._var_names

    @cached_property
    def shape(self):
        return len(self._obs_names), len(self._var_names)

    def to_dict(self):
        prepare_adata = {}

        prepare_adata["X"] = _to_memory(self.X)

        if "uns" in self._attrs_keys:
            prepare_adata["uns"] = self.uns

        for attr in ("obs", "var"):
            if attr in self._attrs_keys:
                prepare_adata[attr] = getattr(self, attr)

        for attr in ("obsm", "varm", "obsp", "varp", "layers"):
            if attr in self._attrs_keys:
                prepare_adata[attr] = {}
                get_attr = getattr(self, attr)
                for key in self._attrs_keys[attr]:
                    prepare_adata[attr][key] = _to_memory(get_attr[key])

        if "raw" in self._attrs_keys:
            prepare_adata["raw"] = self.raw.to_dict()

        return prepare_adata

    def to_memory(self):
        adata = AnnData(**self.to_dict())
        return adata


class AnnDataAccessorSubset(_AnnDataAttrsMixin):
    def __init__(self, storage, indices, attrs_keys, obs_names, var_names, ref_shape):
        self.storage = storage
        self.indices = indices

        self._attrs_keys = attrs_keys
        self._obs_names, self._var_names = obs_names, var_names

        self._ref_shape = ref_shape

    def __getitem__(self, index: Index):
        """Access a subset of the underlying AnnData object."""
        oidx, vidx = _normalize_indices(index, self._obs_names, self._var_names)
        new_obs_names, new_var_names = self._obs_names[oidx], self._var_names[vidx]
        if self.indices is not None:
            oidx = _resolve_idx(self.indices[0], oidx, self._ref_shape[0])
            vidx = _resolve_idx(self.indices[1], vidx, self._ref_shape[1])
        return type(self)(
            self.storage,
            (oidx, vidx),
            self._attrs_keys,
            new_obs_names,
            new_var_names,
            self._ref_shape,
        )

    def __repr__(self):
        """Description of the object."""
        n_obs, n_vars = self.shape
        descr = f"{type(self).__name__} object with n_obs × n_vars = {n_obs} × {n_vars}"
        for attr, keys in self._attrs_keys.items():
            descr += f"\n  {attr}: {keys}"
        return descr

    @cached_property
    def raw(self):
        if "raw" not in self._attrs_keys:
            return None
        prepare_indices = None
        if self.indices is not None:
            oidx = self.indices[0]
            if oidx != slice(None):
                prepare_indices = oidx, slice(None)
        return AnnDataRawAccessor(
            self.storage["raw"],
            prepare_indices,
            None,
            self._obs_names,
            None,
            self._ref_shape[0],
        )


class AnnDataRawAccessor(AnnDataAccessorSubset):
    def __init__(
        self, storage_raw, indices, attrs_keys, obs_names, var_names, ref_shape
    ):
        var_raw = storage_raw["var"]

        if var_names is None:
            var_names = _safer_read_index(var_raw)

        if isinstance(ref_shape, int):
            ref_shape = ref_shape, len(var_names)
        elif isinstance(ref_shape, tuple) and len(ref_shape) < 2:
            ref_shape = ref_shape[0], len(var_names)

        if attrs_keys is None:
            attrs_keys = {}
            if isinstance(var_raw, (h5py.Dataset, zarr.Array)):
                attrs_keys["var"] = list(var_raw.dtype.fields.keys())
            else:
                # for some reason list(var_raw.keys()) is very slow for zarr
                # maybe also directly get keys from the underlying mapper
                attrs_keys["var"] = [key for key in var_raw]
            if "varm" in storage_raw:
                varm_keys_raw = [key for key in storage_raw["varm"]]
                if len(varm_keys_raw) > 0:
                    attrs_keys["varm"] = varm_keys_raw

        super().__init__(
            storage_raw, indices, attrs_keys, obs_names, var_names, ref_shape
        )

    @property
    def raw(self):
        raise AttributeError


class AnnDataAccessor(_AnnDataAttrsMixin):
    """Cloud-backed AnnData."""

    def __init__(
        self,
        connection: Union[OpenFile, None],
        storage: StorageType,
        filename: str,
    ):
        self._conn = connection
        self.storage = storage

        self._attrs_keys = registry.keys(self.storage)

        self._name = filename

        self._obs_names = _safer_read_index(self.storage["obs"])  # type: ignore
        self._var_names = _safer_read_index(self.storage["var"])  # type: ignore

    def close(self):
        """Closes the connection."""
        if self._conn is not None:
            self.storage.close()
            self._conn.close()

    def __del__(self):
        """Closes the connection on deletion."""
        self.close()

    def __getitem__(self, index: Index) -> AnnDataAccessorSubset:
        """Access a subset of the underlying AnnData object."""
        oidx, vidx = _normalize_indices(index, self._obs_names, self._var_names)
        new_obs_names, new_var_names = self._obs_names[oidx], self._var_names[vidx]
        return AnnDataAccessorSubset(
            self.storage,
            (oidx, vidx),
            self._attrs_keys,
            new_obs_names,
            new_var_names,
            self.shape,
        )

    def __repr__(self):
        """Description of the AnnDataAccessor object."""
        n_obs, n_vars = self.shape
        descr = f"AnnDataAccessor object with n_obs × n_vars = {n_obs} × {n_vars}"
        descr += f"\n  constructed for the AnnData object {self._name}"
        for attr, keys in self._attrs_keys.items():
            descr += f"\n    {attr}: {keys}"
        return descr

    @cached_property
    def raw(self):
        if "raw" not in self._attrs_keys:
            return None
        return AnnDataRawAccessor(
            self.storage["raw"], None, None, self._obs_names, None, self.shape[0]
        )


@dataclass
class BackedAccessor:
    """h5py.File or zarr.Group accessor."""

    connection: OpenFile
    """The connection."""
    storage: StorageType
    """The storage access."""


def backed_access(
    file_or_filepath: Union[File, Path]
) -> Union[AnnDataAccessor, BackedAccessor]:
    if isinstance(file_or_filepath, File):
        filepath = filepath_from_file(file_or_filepath)
    else:
        filepath = file_or_filepath
    name = filepath.name

    if filepath.suffix in (".h5", ".hdf5", ".h5ad"):
        conn, storage = registry.open("h5py", filepath)
    elif filepath.suffix in (".zarr", ".zrad"):
        conn, storage = registry.open("zarr", filepath)
    else:
        raise ValueError(
            "file should have .h5, .hdf5, .h5ad, .zarr or .zrad suffix, not"
            f" {filepath.suffix}."
        )

    if filepath.suffix in (".h5ad", ".zrad"):
        return AnnDataAccessor(conn, storage, name)
    else:
        if get_spec(storage).encoding_type == "anndata":
            return AnnDataAccessor(conn, storage, name)
        else:
            return BackedAccessor(conn, storage)



import builtins
from typing import Iterable, List, NamedTuple, Optional, Union

import pandas as pd
from django.core.exceptions import FieldDoesNotExist
from django.db.models import Manager, QuerySet
from lamin_utils import logger
from lamin_utils._lookup import Lookup
from lamin_utils._search import search as base_search
from lamindb_setup.dev._docs import doc_args
from lnschema_core import Registry
from lnschema_core.types import ListLike, StrField

from lamindb._utils import attach_func_to_class_method

from . import _TESTING
from ._from_values import get_or_create_records

IPYTHON = getattr(builtins, "__IPYTHON__", False)


class ValidationError(Exception):
    pass


def init_self_from_db(self: Registry, existing_record: Registry):
    new_args = [
        getattr(existing_record, field.attname) for field in self._meta.concrete_fields
    ]
    super(self.__class__, self).__init__(*new_args)
    self._state.adding = False  # mimic from_db
    self._state.db = "default"


def validate_required_fields(orm: Registry, kwargs):
    required_fields = {
        k.name for k in orm._meta.fields if not k.null and k.default is None
    }
    required_fields_not_passed = {k: None for k in required_fields if k not in kwargs}
    kwargs.update(required_fields_not_passed)
    missing_fields = [
        k for k, v in kwargs.items() if v is None and k in required_fields
    ]
    if missing_fields:
        raise TypeError(f"{missing_fields} are required.")


def suggest_objects_with_same_name(orm: Registry, kwargs) -> Optional[str]:
    if kwargs.get("name") is None:
        return None
    else:
        results = orm.search(kwargs["name"])
        if results.shape[0] == 0:
            return None

        # subset results to those with at least 0.90 levensteihn distance
        results = results.loc[results.__ratio__ >= 90]

        # test for exact match
        if len(results) > 0:
            if results.index[0] == kwargs["name"]:
                return "object-with-same-name-exists"
            else:
                s = "" if results.shape[0] == 1 else "s"
                it = "it" if results.shape[0] == 1 else "one of them"
                msg = (
                    f"record{s} with similar name{s} exist! did you mean to load {it}?"
                )
                if IPYTHON:
                    from IPython.display import display

                    logger.warning(f"{msg}")
                    display(results)
                else:
                    logger.warning(f"{msg}\n{results}")
    return None


def __init__(orm: Registry, *args, **kwargs):
    if not args:
        validate_required_fields(orm, kwargs)
        from .dev._settings import settings

        if settings.upon_create_search_names:
            result = suggest_objects_with_same_name(orm, kwargs)
            if result == "object-with-same-name-exists":
                if "version" in kwargs:
                    version_comment = " and version"
                    existing_record = orm.filter(
                        name=kwargs["name"], version=kwargs["version"]
                    ).one_or_none()
                else:
                    version_comment = ""
                    existing_record = orm.filter(name=kwargs["name"]).one()
                if existing_record is not None:
                    logger.success(
                        f"loaded {orm.__class__.__name__} record with exact same"
                        f" name{version_comment}: '{kwargs['name']}'"
                    )
                    init_self_from_db(orm, existing_record)
                    return None
        super(Registry, orm).__init__(**kwargs)
    elif len(args) != len(orm._meta.concrete_fields):
        raise ValueError("please provide keyword arguments, not plain arguments")
    else:
        # object is loaded from DB (**kwargs could be omitted below, I believe)
        super(Registry, orm).__init__(*args, **kwargs)


# from_values doesn't apply for QuerySet or Manager
@classmethod  # type:ignore
@doc_args(Registry.from_values.__doc__)
def from_values(
    cls, values: ListLike, field: Optional[StrField] = None, **kwargs
) -> List["Registry"]:
    """{}"""
    from_bionty = True if cls.__module__.startswith("lnschema_bionty.") else False
    field_str = get_default_str_field(cls, field=field)
    return get_or_create_records(
        iterable=values,
        field=getattr(cls, field_str),
        from_bionty=from_bionty,
        **kwargs,
    )


# From: https://stackoverflow.com/a/37648265
def _order_queryset_by_ids(queryset: QuerySet, ids: Iterable):
    from django.db.models import Case, When

    preserved = Case(*[When(pk=pk, then=pos) for pos, pk in enumerate(ids)])
    return queryset.filter(pk__in=ids).order_by(preserved)


def _search(
    cls,
    string: str,
    *,
    field: Optional[Union[StrField, List[StrField]]] = None,
    limit: Optional[int] = 20,
    return_queryset: bool = False,
    case_sensitive: bool = False,
    synonyms_field: Optional[StrField] = "synonyms",
) -> Union["pd.DataFrame", "QuerySet"]:
    queryset = _queryset(cls)
    orm = queryset.model

    def _search_single_field(
        string: str,
        field: Optional[StrField],
        synonyms_field: Optional[StrField] = "synonyms",
    ) -> "pd.DataFrame":
        field = get_default_str_field(orm=orm, field=field)

        try:
            orm._meta.get_field(synonyms_field)
            synonyms_field_exists = True
        except FieldDoesNotExist:
            synonyms_field_exists = False

        if synonyms_field is not None and synonyms_field_exists:
            df = pd.DataFrame(queryset.values("id", field, synonyms_field))
        else:
            df = pd.DataFrame(queryset.values("id", field))

        return base_search(
            df=df,
            string=string,
            field=field,
            limit=limit,
            synonyms_field=str(synonyms_field),
            case_sensitive=case_sensitive,
        )

    # search in both key and description fields for file
    if orm._meta.model.__name__ == "File" and field is None:
        field = ["key", "description"]

    if not isinstance(field, List):
        field = [field]

    results = []
    for fd in field:
        result_field = _search_single_field(
            string=string, field=fd, synonyms_field=synonyms_field
        )
        results.append(result_field)
        # turn off synonyms search after the 1st field
        synonyms_field = None

    if len(results) > 1:
        result = (
            pd.concat([r.reset_index() for r in results], join="outer")
            .drop(columns=["index"], errors="ignore")
            .set_index("id")
        )
    else:
        result = results[0]

    # remove results that have __ratio__ 0
    if "__ratio__" in result.columns:
        result = result[result["__ratio__"] > 0].sort_values(
            "__ratio__", ascending=False
        )
        # move the __ratio__ to be the last column
        result["__ratio__"] = result.pop("__ratio__")

    if return_queryset:
        return _order_queryset_by_ids(queryset, result.reset_index()["id"])
    else:
        return result.fillna("")


@classmethod  # type: ignore
@doc_args(Registry.search.__doc__)
def search(
    cls,
    string: str,
    *,
    field: Optional[StrField] = None,
    limit: Optional[int] = 20,
    return_queryset: bool = False,
    case_sensitive: bool = False,
    synonyms_field: Optional[StrField] = "synonyms",
) -> Union["pd.DataFrame", "QuerySet"]:
    """{}"""
    return _search(
        cls=cls,
        string=string,
        field=field,
        return_queryset=return_queryset,
        limit=limit,
        case_sensitive=case_sensitive,
        synonyms_field=synonyms_field,
    )


def _lookup(
    cls, field: Optional[StrField] = None, return_field: Optional[StrField] = None
) -> NamedTuple:
    """{}"""
    queryset = _queryset(cls)
    field = get_default_str_field(orm=queryset.model, field=field)

    return Lookup(
        records=queryset,
        values=[i.get(field) for i in queryset.values()],
        tuple_name=cls.__class__.__name__,
        prefix="ln",
    ).lookup(
        return_field=get_default_str_field(orm=queryset.model, field=return_field)
        if return_field is not None
        else None
    )


@classmethod  # type: ignore
@doc_args(Registry.lookup.__doc__)
def lookup(
    cls, field: Optional[StrField] = None, return_field: Optional[StrField] = None
) -> NamedTuple:
    """{}"""
    return _lookup(cls=cls, field=field, return_field=return_field)


def get_default_str_field(
    orm: Union[Registry, QuerySet, Manager],
    *,
    field: Optional[Union[str, StrField]] = None,
) -> str:
    """Get the 1st char or text field from the orm."""
    if isinstance(orm, (QuerySet, Manager)):
        orm = orm.model
    model_field_names = [i.name for i in orm._meta.fields]

    # set default field
    if field is None:
        if orm._meta.model.__name__ == "Run":
            field = orm._meta.get_field("created_at")
        elif orm._meta.model.__name__ == "User":
            field = orm._meta.get_field("handle")
        elif "name" in model_field_names:
            # by default use the name field
            field = orm._meta.get_field("name")
        else:
            # first char or text field that doesn't contain "id"
            for i in orm._meta.fields:
                if "id" in i.name:
                    continue
                if i.get_internal_type() in {"CharField", "TextField"}:
                    field = i
                    break

        # no default field can be found
        if field is None:
            raise ValueError(
                "please pass a Registry string field, e.g., `CellType.name`!"
            )
        else:
            field = field.name  # type:ignore
    if not isinstance(field, str):
        try:
            field = field.field.name
        except AttributeError:
            raise TypeError(
                "please pass a Registry string field, e.g., `CellType.name`!"
            )

    return field


def _queryset(cls: Union[Registry, QuerySet, Manager]) -> QuerySet:
    queryset = cls.all() if isinstance(cls, QuerySet) else cls.objects.all()
    return queryset


METHOD_NAMES = [
    "__init__",
    "search",
    "lookup",
    "from_values",
]

if _TESTING:  # type: ignore
    from inspect import signature

    SIGS = {
        name: signature(getattr(Registry, name))
        for name in METHOD_NAMES
        if not name.startswith("__")
    }

for name in METHOD_NAMES:
    attach_func_to_class_method(name, Registry, globals())


@classmethod  # type: ignore
def __get_schema_name__(cls) -> str:
    schema_module_name = cls.__module__.split(".")[0]
    schema_name = schema_module_name.replace("lnschema_", "")
    return schema_name


@classmethod  # type: ignore
def __get_name_with_schema__(cls) -> str:
    schema_name = cls.__get_schema_name__()
    return f"{schema_name}.{cls.__name__}"


def select_backward(cls, **expressions):
    logger.warning("select() is deprecated! please use: Registry.filter()")
    return cls.filter(**expressions)


@classmethod  # type: ignore
def select(cls, **expressions):
    return select_backward(cls, **expressions)


setattr(Registry, "__get_schema_name__", __get_schema_name__)
setattr(Registry, "__get_name_with_schema__", __get_name_with_schema__)
setattr(Registry, "select", select)  # backward compat



import shutil

import pytest
from lamin_utils import logger

import lamindb_setup


def pytest_sessionstart(session: pytest.Session):
    lamindb_setup.init(
        storage="./default_storage",
        schema="bionty",
        name="lamindb-setup-unit-tests",
    )


def pytest_sessionfinish(session: pytest.Session):
    logger.set_verbosity(1)
    lamindb_setup.delete("lamindb-setup-unit-tests", force=True)
    shutil.rmtree("./default_storage")



# Generated by Django 4.2.2 on 2023-07-28 13:22

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("lnschema_core", "0012_remove_label_ref_id_remove_label_ref_orm_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="feature",
            name="labels_orm",
        ),
        migrations.RemoveField(
            model_name="feature",
            name="labels_schema",
        ),
        migrations.AddField(
            model_name="feature",
            name="registries",
            field=models.CharField(db_index=True, default=None, max_length=128, null=True),
        ),
        migrations.AlterUniqueTogether(
            name="label",
            unique_together=set(),
        ),
        migrations.CreateModel(
            name="FileLabel",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("feature", models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, to="lnschema_core.feature")),
                ("file", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="lnschema_core.file")),
                ("label", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="lnschema_core.label")),
            ],
            options={
                "unique_together": {("file", "label")},
            },
        ),
        migrations.RunSQL("CREATE TABLE lnschema_core_filelabel_tmp (id BIGINT, file_id TEXT, label_id TEXT)"),
        migrations.RunSQL("INSERT INTO lnschema_core_filelabel_tmp (id, file_id, label_id) SELECT id, file_id, label_id from lnschema_core_file_labels"),
        migrations.RemoveField(
            model_name="file",
            name="labels",
        ),
        migrations.AddField(
            model_name="file",
            name="labels",
            field=models.ManyToManyField(related_name="files", through="lnschema_core.FileLabel", to="lnschema_core.label"),
        ),
        migrations.RunSQL("INSERT INTO lnschema_core_filelabel (id, file_id, label_id) SELECT id, file_id, label_id from lnschema_core_filelabel_tmp"),
        migrations.RunSQL("DROP TABLE lnschema_core_filelabel_tmp"),
        migrations.RemoveField(
            model_name="label",
            name="feature",
        ),
        migrations.RemoveField(
            model_name="featureset",
            name="ref_field",
        ),
        migrations.RenameField(
            model_name="featureset",
            old_name="ref_orm",
            new_name="ref_field",
        ),
        migrations.RemoveField(
            model_name="featureset",
            name="ref_schema",
        ),
        migrations.AlterField(
            model_name="featureset",
            name="ref_field",
            field=models.CharField(db_index=True, max_length=128),
        ),
    ]



import pytest

import lamindb as ln


def test_ulabel():
    with pytest.raises(ValueError) as error:
        ln.ULabel(x=1)
    assert (
        error.exconly()
        == "ValueError: Only name, description, reference, reference_type are valid keyword arguments"  # noqa
    )

    with pytest.raises(ValueError) as error:
        ln.ULabel(1)
    assert error.exconly() == "ValueError: Only one non-keyword arg allowed"



import os
from pathlib import Path

from pydantic import BaseSettings

# user_config_dir in appdirs is weird on MacOS!
# hence, let's take home/.lamin
settings_dir = Path.home() / ".lamin"
settings_dir.mkdir(parents=True, exist_ok=True)


def get_settings_file_name_prefix():
    if "LAMIN_ENV" in os.environ:
        if os.environ["LAMIN_ENV"] != "prod":
            return f"{os.environ['LAMIN_ENV']}--"
    return ""


def current_instance_settings_file():
    return settings_dir / f"{get_settings_file_name_prefix()}current_instance.env"


def current_user_settings_file():
    return settings_dir / f"{get_settings_file_name_prefix()}current_user.env"


def instance_settings_file(name: str, owner: str):
    return (
        settings_dir / f"{get_settings_file_name_prefix()}instance--{owner}--{name}.env"
    )


def user_settings_file_email(email: str):
    return settings_dir / f"{get_settings_file_name_prefix()}user--{email}.env"


def user_settings_file_handle(handle: str):
    return settings_dir / f"{get_settings_file_name_prefix()}user--{handle}.env"


class InstanceSettingsStore(BaseSettings):
    owner: str
    name: str
    storage_root: str
    storage_region: str  # should be Optional, but we use types for instantiating
    db: str  # should be Optional, but we use types for instantiating
    schema_str: str  # should be Optional, but we use types for instantiating
    id: str  # should be Optional, but we use types for instantiating

    class Config:
        env_prefix = "lamindb_instance_"
        env_file = ".env"


class UserSettingsStore(BaseSettings):
    email: str
    password: str
    access_token: str
    id: str
    uuid: str
    handle: str
    name: str

    class Config:
        env_prefix = "lamin_user_"
        env_file = ".env"



"""Schema tools & overview.

Guide: :doc:`/schemas`

You can access mounted schema modules with domain-specific entities via
available via `ln.schema.<module>.<entity>`.

However, we recommend to import schema modules, e.g., like `import
lnschema_bionty as bt`.

.. autosummary::
   :toctree: .

   graph
   view

"""
import importlib as _importlib

from lamindb_setup import settings as _settings
from lamindb_setup._init_instance import reload_schema_modules as _reload_schema_modules

from .. import _INSTANCE_SETUP

if _INSTANCE_SETUP:
    _reload_schema_modules(_settings.instance)

from ._core import graph, view



from typing import List, Union, overload  # noqa

from lamin_utils import colors, logger
from lnschema_core import Registry


@overload
def delete(
    record: Registry,
) -> None:
    ...  # pragma: no cover


@overload
def delete(
    records: List[Registry],
) -> None:  # type: ignore
    ...  # pragma: no cover


def delete(  # type: ignore
    records: Union[Registry, List[Registry]],
) -> None:
    """Delete metadata records & files.

    Args:
        records: `Union[Registry, List[Registry]]` One or multiple records.

    Returns:
        `None`

    Examples:

        Delete by record:

        >>> experiment = ln.filter(Experiment, id=experiment_id).one()
        >>> ln.delete(experiment)

        Delete files (delete the metadata record and the file in storage):

        >>> file = ln.filter(File, id=file_id).one()
        >>> ln.delete(file)
        >>> # deleting the record occurs automatically
        >>> # you will be asked whether to delete the file in storage
        >>> # for more control, use:
        >>> file.delete(storage=True)

        Bulk delete via QuerySet:

        >>> ln.save(ln.ULabel.from_values(["ULabel1", "ULabel2", "ULabel3"], field="name"))
        >>> queryset = ln.ULabel.filter(name__icontains = "label")
        >>> queryset.list()
        [ULabel(id=o3FY3c5n, name=ULabel2, updated_at=2023-07-19 18:28:16, created_by_id=kmvZDIX9), # noqa
        ULabel(id=Qi3c4utq, name=ULabel3, updated_at=2023-07-19 18:28:16, created_by_id=kmvZDIX9), # noqa
        ULabel(id=CcFPLmpq, name=ULabel1, updated_at=2023-07-19 18:28:16, created_by_id=kmvZDIX9)] # noqa
        >>> queryset.delete()
    """
    logger.warning("for efficient bulk delete, use `queryset.delete` instead")
    if isinstance(records, list):
        records = records
    elif isinstance(records, Registry):
        records = [records]
    for record in records:
        record.delete()
        logger.success(f"deleted {colors.yellow(f'{record}')}")



from typing import Dict, List, Union

from lamin_utils import colors
from lnschema_core.models import Data, Dataset, Feature, File, Registry

from .._from_values import _print_values
from .._query_set import QuerySet
from .._registry import get_default_str_field
from ._feature_manager import dict_related_model_to_related_name


def print_labels(self: Data):
    labels_msg = ""
    for related_model, related_name in dict_related_model_to_related_name(
        self.__class__
    ).items():
        if related_name in {"feature_sets", "files", "input_of"}:
            continue
        labels = self.__getattribute__(related_name)
        if labels.exists():
            n = labels.count()
            field = get_default_str_field(labels)
            print_values = _print_values(labels.list(field), n=10)
            labels_msg += f"  🏷️ {related_name} ({n}, {colors.italic(related_model)}): {print_values}\n"  # noqa
    if len(labels_msg) > 0:
        return f"{colors.green('Labels')}:\n{labels_msg}"
    else:
        return ""


class LabelManager:
    """Label manager (:attr:`~lamindb.dev.Data.labels`).

    See :class:`~lamindb.dev.Data` for more information.
    """

    def __init__(self, host: Union[File, Dataset]):
        self._host = host

    def __repr__(self) -> str:
        msg = print_labels(self._host)
        if len(msg) > 0:
            return msg
        else:
            return "no linked labels"

    def add(
        self,
        records: Union[Registry, List[Registry], QuerySet],
        feature: Feature,
    ) -> None:
        """Add one or several labels and associate them with a feature.

        Args:
            records: Label records to add.
            feature: Feature under which to group the labels.
            field: Field to parse iterable with from_values.
        """
        from ._data import add_labels

        return add_labels(self._host, records=records, feature=feature)

    def get(
        self,
        feature: Feature,
        mute: bool = False,
        flat_names: bool = False,
    ) -> Union[QuerySet, Dict[str, QuerySet], List]:
        """Get labels given a feature.

        Args:
            feature: Feature under which labels are grouped.
            mute: Show no logging.
            flat_names: Flatten list to names rather than returning records.
        """
        from ._data import get_labels

        return get_labels(self._host, feature=feature, mute=mute, flat_names=flat_names)



def attach_func_to_class_method(func_name, cls, globals):
    implementation = globals[func_name]
    target = getattr(cls, func_name)
    # assigning the original class definition docstring
    # to the implementation only has an effect for regular methods
    # not for class methods
    # this is why we need @doc_args for class methods
    implementation.__doc__ = target.__doc__
    setattr(cls, func_name, implementation)



import shutil
from inspect import signature
from pathlib import Path

import anndata as ad
import lamindb_setup
import lnschema_bionty as lb
import numpy as np
import pandas as pd
import pytest
from django.db.models.deletion import ProtectedError
from lamindb_setup.dev.upath import CloudPath, LocalPathClasses, UPath

import lamindb as ln
from lamindb import _file
from lamindb._file import (
    check_path_is_child_of_root,
    get_hash,
    get_relative_path_to_directory,
    process_data,
)
from lamindb.dev.storage._zarr import write_adata_zarr
from lamindb.dev.storage.file import (
    AUTO_KEY_PREFIX,
    auto_storage_key_from_id_suffix,
    delete_storage,
    extract_suffix_from_path,
    load_to_memory,
    read_fcs,
    read_tsv,
)

# how do we properly abstract out the default storage variable?
# currently, we're only mocking it through `default_storage` as
# set in conftest.py

ln.settings.verbosity = 3
lb.settings.species = "human"

df = pd.DataFrame({"feat1": [1, 2], "feat2": [3, 4]})

adata = ad.AnnData(
    X=np.array([[1, 2, 3], [4, 5, 6]]),
    obs=dict(feat1=["A", "B"]),
    var=pd.DataFrame(index=["MYC", "TCF7", "GATA1"]),
    obsm=dict(X_pca=np.array([[1, 2], [3, 4]])),
)


def test_signatures():
    # this seems currently the easiest and most transparent
    # way to test violations of the signature equality
    # the MockORM class is needed to get inspect.signature
    # to work
    class Mock:
        pass

    # class methods
    class_methods = ["view_tree", "from_dir", "from_df", "from_anndata"]
    for name in class_methods:
        setattr(Mock, name, getattr(_file, name))
        assert signature(getattr(Mock, name)) == _file.SIGS.pop(name)
    # methods
    for name, sig in _file.SIGS.items():
        assert signature(getattr(_file, name)) == sig


def test_is_new_version_of_versioned_file():
    # attempt to create a file with an invalid version
    with pytest.raises(ValueError) as error:
        file = ln.File(df, description="test", version=0)
    assert (
        error.exconly()
        == "ValueError: `version` parameter must be `None` or `str`, e.g., '0.1', '1',"
        " '2', etc."
    )

    # create a versioned file
    file = ln.File(df, description="test", version="1")
    assert file.version == "1"

    assert file.path.exists()  # because of cache file already exists
    file.save()
    assert file.path.exists()

    with pytest.raises(ValueError) as error:
        file_v2 = ln.File(adata, is_new_version_of=file, version="1")
    assert error.exconly() == "ValueError: Please increment the previous version: '1'"

    # create new file from old file
    file_v2 = ln.File(adata, is_new_version_of=file)
    assert file_v2.id[:18] == file.id[:18]  # stem_id
    assert file.version == "1"
    assert file.initial_version_id is None  # initial file has initial_version_id None
    assert file_v2.initial_version_id == file.id
    assert file_v2.version == "2"
    assert file_v2.key is None
    assert file_v2.description == "test"

    file_v2.save()
    assert file_v2.path.exists()

    # create new file from newly versioned file
    df.iloc[0, 0] = 0
    file_v3 = ln.File(df, description="test1", is_new_version_of=file_v2)
    assert file_v3.id[:18] == file.id[:18]  # stem_id
    assert file_v3.initial_version_id == file.id
    assert file_v3.version == "3"
    assert file_v3.description == "test1"

    with pytest.raises(TypeError) as error:
        ln.File(df, description="test1a", is_new_version_of=ln.Transform())
    assert error.exconly() == "TypeError: is_new_version_of has to be of type ln.File"

    # test that reference file cannot be deleted
    with pytest.raises(ProtectedError):
        file.delete(storage=True)
    file_v2.delete(storage=True)
    file_v3.delete(storage=True)
    file.delete(storage=True)

    # extra kwargs
    with pytest.raises(ValueError):
        ln.File(df, description="test1b", extra_kwarg="extra")

    # > 1 args
    with pytest.raises(ValueError) as error:
        ln.File(df, df)
    assert error.exconly() == "ValueError: Only one non-keyword arg allowed: data"

    # AUTO_KEY_PREFIX
    with pytest.raises(ValueError) as error:
        ln.File(df, key=".lamindb/test_df.parquet")
    assert error.exconly() == "ValueError: Key cannot start with .lamindb/"


def test_is_new_version_of_unversioned_file():
    # unversioned file
    file = ln.File(df, description="test2")
    assert file.initial_version_id is None
    assert file.version is None

    # what happens if we don't save the old file?
    # add a test for it!
    file.save()

    # create new file from old file
    new_file = ln.File(adata, is_new_version_of=file)
    assert new_file.id[:18] == file.id[:18]  # stem_id
    assert file.version == "1"
    assert file.initial_version is None
    assert new_file.initial_version_id == file.id
    assert new_file.version == "2"
    assert new_file.description == file.description

    file.delete(storage=True)


# also test legacy name parameter (got removed by description)
def test_create_from_dataframe():
    file = ln.File(df, description="test1")
    assert file.description == "test1"
    assert file.key is None
    assert file.accessor == "DataFrame"
    assert hasattr(file, "_local_filepath")
    file.save()
    # can't do backed
    with pytest.raises(ValueError):
        file.backed()
    # check that the local filepath has been cleared
    assert not hasattr(file, "_local_filepath")
    file.delete(storage=True)


def test_create_from_dataframe_using_from_df():
    description = "my description"
    file = ln.File.from_df(df, description=description)
    assert file._feature_sets == {}
    with pytest.raises(ValueError):
        file.features["columns"]
    ln.save(ln.Feature.from_df(df))
    file = ln.File.from_df(df, description=description)
    ln.Modality(name="random").save()
    modalities = ln.Modality.lookup()
    file = ln.File.from_df(df, description=description, modality=modalities.random)
    # mere access test right now
    file.features["columns"]
    assert file.description == description
    assert file.key is None
    assert file.accessor == "DataFrame"
    assert hasattr(file, "_local_filepath")
    file.save()
    # check that the local filepath has been cleared
    assert not hasattr(file, "_local_filepath")
    feature_set_queried = file.feature_sets.get()  # exactly one
    assert feature_set_queried.modality == modalities.random
    feature_list_queried = ln.Feature.filter(feature_sets=feature_set_queried).list()
    feature_list_queried = [feature.name for feature in feature_list_queried]
    assert set(feature_list_queried) == set(df.columns)
    feature_set_queried.delete()
    file.delete(storage=True)
    ln.Feature.filter(name__in=["feat1", "feat2"]).delete()


def test_create_from_anndata_in_memory():
    ln.save(lb.Gene.from_values(adata.var.index, "symbol"))
    ln.save(ln.Feature.from_df(adata.obs))
    ln.Modality(name="random").save()
    modalities = ln.Modality.lookup()
    file = ln.File.from_anndata(
        adata, description="test", field=lb.Gene.symbol, modality=modalities.random
    )
    assert file.accessor == "AnnData"
    assert hasattr(file, "_local_filepath")
    file.save()
    # check that the local filepath has been cleared
    assert not hasattr(file, "_local_filepath")
    feature_sets_queried = file.feature_sets.all()
    features_queried = ln.Feature.filter(feature_sets__in=feature_sets_queried).all()
    assert set(features_queried.list("name")) == set(adata.obs.columns)
    genes_queried = lb.Gene.filter(feature_sets__in=feature_sets_queried).all()
    assert set(genes_queried.list("symbol")) == set(adata.var.index)
    feature_sets_queried.delete()
    features_queried.delete()
    genes_queried.delete()
    file.delete(storage=True)


@pytest.mark.parametrize(
    "data", [adata, "s3://lamindb-test/scrnaseq_pbmc68k_tiny.h5ad"]
)
def test_create_from_anndata_in_storage(data):
    if isinstance(data, ad.AnnData):
        filepath = Path("./default_storage/test_adata.h5ad")
        data.write(filepath)
    else:
        previous_storage = ln.setup.settings.storage.root_as_str
        ln.settings.storage = "s3://lamindb-test"
        filepath = data
    file = ln.File.from_anndata(filepath, field=lb.Gene.symbol)
    assert file.accessor == "AnnData"
    assert hasattr(file, "_local_filepath")
    file.save()
    # check that the local filepath has been cleared
    assert not hasattr(file, "_local_filepath")
    if isinstance(data, ad.AnnData):
        filepath.unlink()
    else:
        ln.settings.storage = previous_storage


@pytest.fixture(
    scope="module",
    params=[
        # tuple of isin_existing_storage, path, suffix
        (True, "./default_storage/", ".csv"),
        (True, "./default_storage/", ""),
        (True, "./registered_storage/", ".csv"),
        (True, "./registered_storage/", ""),
        (False, "./nonregistered_storage/", ".csv"),
        (False, "./nonregistered_storage/", ""),
    ],
)
def get_test_filepaths(request):  # -> Tuple[bool, Path, Path, Path, str]
    isin_existing_storage: bool = request.param[0]
    root_dir: Path = Path(request.param[1])
    suffix: str = request.param[2]
    if isin_existing_storage:
        # ensure that it's actually registered
        if ln.Storage.filter(root=root_dir.resolve().as_posix()).one_or_none() is None:
            ln.Storage(root=root_dir.resolve().as_posix(), type="local").save()
    test_dir = root_dir / "my_dir/"
    test_dir.mkdir(parents=True)
    test_filepath = test_dir / f"my_file{suffix}"
    test_filepath.write_text(str(test_filepath))
    # create a duplicated file
    test_filepath1 = test_dir / f"my_file1{suffix}"
    test_filepath1.write_text(str(test_filepath))
    # create a non-duplicated file
    test_filepath2 = test_dir / f"my_file2{suffix}"
    test_filepath2.write_text(str(test_filepath2))
    # return a boolean indicating whether test filepath is in default storage
    # and the test filepath
    yield (isin_existing_storage, root_dir, test_dir, test_filepath, suffix)
    shutil.rmtree(test_dir)


# this tests the basic (non-provenance-related) metadata
@pytest.mark.parametrize("key", [None, "my_new_dir/my_file.csv", "nosuffix"])
@pytest.mark.parametrize("description", [None, "my description"])
def test_create_from_local_filepath(get_test_filepaths, key, description):
    isin_existing_storage = get_test_filepaths[0]
    root_dir = get_test_filepaths[1]
    test_filepath = get_test_filepaths[3]
    suffix = get_test_filepaths[4]
    # this test insufficient information being provided
    if key is None and not isin_existing_storage and description is None:
        # this can fail because ln.track() might set a global run context
        # in that case, the File would have a run that's not None and the
        # error below wouldn't be thrown
        with pytest.raises(ValueError) as error:
            file = ln.File(test_filepath, key=key, description=description)
        assert (
            error.exconly()
            == "ValueError: Pass one of key, run or description as a parameter"
        )
        return None
    elif key is not None and isin_existing_storage:
        inferred_key = get_relative_path_to_directory(
            path=test_filepath, directory=root_dir
        ).as_posix()
        with pytest.raises(ValueError) as error:
            file = ln.File(test_filepath, key=key, description=description)
        assert (
            error.exconly()
            == f"ValueError: The path '{test_filepath}' is already in registered"
            " storage"
            f" '{root_dir.resolve().as_posix()}' with key '{inferred_key}'\nYou"
            f" passed conflicting key '{key}': please move the file before"
            " registering it."
        )
        return None
    else:
        file = ln.File(test_filepath, key=key, description=description)
        assert file._state.adding  # make sure that this is a new file in the db
    assert (
        file.description is None
        if description is None
        else file.description == description
    )
    assert file.suffix == suffix
    if key is None:
        assert (
            file.key == f"my_dir/my_file{suffix}"
            if isin_existing_storage
            else file.key is None
        )
        if isin_existing_storage:
            assert file.storage.root == root_dir.resolve().as_posix()
        else:
            assert file.storage.root == lamindb_setup.settings.storage.root_as_str
    else:
        assert file.key == key
        assert file.storage.root == lamindb_setup.settings.storage.root_as_str

    # test that the file didn't move
    if isin_existing_storage and key is None:
        assert str(test_filepath.resolve()) == str(file.path)

    file.save()
    assert file.path.exists()

    # only delete from storage if a file copy took place
    delete_from_storage = str(test_filepath.resolve()) != str(file.path)
    file.delete(storage=delete_from_storage)


def test_local_path_load():
    local_filepath = ln.dev.datasets.anndata_file_pbmc68k_test().resolve()

    file = ln.File(local_filepath, description="test")
    assert local_filepath == file._local_filepath
    assert local_filepath == file.path
    assert local_filepath == file.stage()

    adata = ad.read(local_filepath)
    file = ln.File(adata, description="test")
    assert file._memory_rep is adata
    assert file.load() is adata
    assert file._local_filepath.resolve() == file.stage() == file.path


ERROR_MESSAGE = """\
ValueError: Currently don't support tracking folders outside one of the storage roots:
"""


@pytest.mark.parametrize("key", [None, "my_new_folder"])
def test_from_dir(get_test_filepaths, key):
    isin_existing_storage = get_test_filepaths[0]
    # root_dir = get_test_filepaths[1]
    test_dirpath = get_test_filepaths[2]
    # the directory contains 3 files, two of them are duplicated
    if key is not None and isin_existing_storage:
        with pytest.raises(ValueError) as error:
            ln.File.from_dir(test_dirpath, key=key)
        assert error.exconly().startswith(
            "ValueError: The path"  # The path {data} is already in registered storage
        )
        return None
    else:
        files = ln.File.from_dir(test_dirpath, key=key)
    # we only return the duplicated ones
    hashes = [file.hash for file in files if file.hash is not None]
    assert len(set(hashes)) == len(hashes)
    ln.File.view_tree(test_dirpath)
    # now save
    ln.save(files)
    # now run again, because now we'll have hash-based lookup!
    files = ln.File.from_dir(test_dirpath, key=key)
    assert len(files) == 2
    assert len(set(hashes)) == len(hashes)
    for file in files:
        file.delete(storage=False)


def test_delete(get_test_filepaths):
    test_filepath = get_test_filepaths[3]
    file = ln.File(test_filepath, description="My test file to delete")
    file.save()
    storage_path = file.path
    file.delete(storage=True)
    assert ln.File.filter(description="My test file to delete").first() is None
    assert not Path(storage_path).exists()


# why does this run so long? in particular the first time?
@pytest.mark.parametrize(
    "filepath_str",
    ["s3://lamindb-ci/test-data/test.parquet", "s3://lamindb-ci/test-data/test.csv"],
)
@pytest.mark.parametrize("skip_check_exists", [False, True])
@pytest.mark.parametrize("skip_size_and_hash", [False, True])
def test_create_small_file_from_remote_path(
    filepath_str, skip_check_exists, skip_size_and_hash
):
    ln.settings.upon_file_create_skip_size_hash = skip_size_and_hash
    file = ln.File(
        filepath_str,
        skip_check_exists=skip_check_exists,
    )
    file.save()
    # test stage()
    file_from_local = ln.File(file.stage(), description="test")
    # test hash equivalency when computed on local machine
    if not skip_size_and_hash:
        assert file_from_local.hash == file.hash
        assert file_from_local.hash_type == "md5"
        assert file.hash_type == "md5"
    assert file.path.as_posix() == filepath_str
    assert file.load().iloc[0].tolist() == [
        0,
        "Abingdon island giant tortoise",
        "Chelonoidis abingdonii",
        106734,
        "ASM359739v1",
        "GCA_003597395.1",
        "Full genebuild",
        "-",
        "-",
    ]
    file.delete(storage=False)
    ln.settings.upon_file_create_skip_size_hash = False


def test_create_big_file_from_remote_path():
    previous_storage = ln.setup.settings.storage.root_as_str
    ln.settings.storage = "s3://lamindb-test"
    filepath_str = "s3://lamindb-test/human_immune.h5ad"
    file = ln.File(filepath_str)
    assert file.key == "human_immune.h5ad"
    assert file.hash.endswith("-2")
    assert file.hash_type == "md5-n"
    ln.settings.storage = previous_storage


# -------------------------------------------------------------------------------------
# Storage
# -------------------------------------------------------------------------------------


def test_extract_suffix_from_path():
    # this is a dataset of path, stem, suffix tuples
    dataset = [
        ("a", "a", ""),
        ("a.txt", "a", ".txt"),
        ("archive.tar.gz", "archive", ".tar.gz"),
        ("directory/file", "file", ""),
        ("d.x.y.z/f.b.c", "f", ".c"),
        ("d.x.y.z/f.a.b.c", "f", ".c"),
        ("logs/date.log.txt", "date", ".txt"),
        ("salmon.merged.gene_counts.tsv", "salmon.merged.gene_counts", ".tsv"),
        ("salmon.merged.gene_counts.tsv.gz", "salmon.merged.gene_counts", ".tsv.gz"),
    ]
    for path, _, suffix in dataset:
        filepath = Path(path)
        assert suffix == extract_suffix_from_path(filepath)


@pytest.mark.parametrize("suffix", [".txt", "", None])
def test_auto_storage_key_from_id_suffix(suffix):
    test_id = "abo389f"
    if suffix is None:
        with pytest.raises(AssertionError):
            auto_storage_key_from_id_suffix(test_id, suffix)
    else:
        assert AUTO_KEY_PREFIX == ".lamindb/"
        storage_key = auto_storage_key_from_id_suffix(test_id, suffix)
        storage_key == f"{AUTO_KEY_PREFIX}{test_id}{suffix}"


def test_storage_root_upath_equivalence():
    storage_root = UPath("s3://lamindb-ci")
    filepath = UPath("s3://lamindb-ci/test-data/Species.csv")
    assert filepath.parents[-1] == storage_root


def test_get_relative_path_to_directory():
    # upath on S3
    upath_root = UPath("s3://lamindb-ci")
    upath_directory1 = UPath("s3://lamindb-ci/test-data")  # no trailing slash
    upath_directory2 = UPath("s3://lamindb-ci/test-data/")  # trailing slash
    upath_file = UPath("s3://lamindb-ci/test-data/test.csv")
    assert (
        "test-data/test.csv"
        == get_relative_path_to_directory(upath_file, upath_root).as_posix()
    )
    assert (
        "test.csv"
        == get_relative_path_to_directory(upath_file, upath_directory1).as_posix()
    )
    assert (
        "test.csv"
        == get_relative_path_to_directory(upath_file, upath_directory2).as_posix()
    )
    # local path
    root = Path("/lamindb-ci")
    upath = Path("/lamindb-ci/test-data/test.csv")
    assert (
        "test-data/test.csv"
        == get_relative_path_to_directory(upath, directory=root).as_posix()
    )
    with pytest.raises(TypeError) as error:
        get_relative_path_to_directory(upath, directory=".")
    assert error.exconly() == "TypeError: Directory not of type Path or UPath"


def test_check_path_is_child_of_root():
    # UPath
    root = UPath("s3://lamindb-ci")
    upath = UPath("s3://lamindb-ci/test-data/test.csv")
    assert check_path_is_child_of_root(upath, root=root)
    upath2 = UPath("s3://lamindb-setup/test-data/test.csv")
    assert not check_path_is_child_of_root(upath2, root=root)
    # local path
    root = Path("/lamindb-ci")
    path = Path("/lamindb-ci/test-data/test.csv")
    assert check_path_is_child_of_root(path, root=root)
    path = Path("/lamindb-other/test-data/test.csv")
    assert not check_path_is_child_of_root(path, root=root)
    # Local & UPath
    root = UPath("s3://lamindb-ci")
    path = Path("/lamindb-ci/test-data/test.csv")
    assert not check_path_is_child_of_root(path, root=root)
    # default root
    assert not check_path_is_child_of_root(path)


def test_serialize_paths():
    fp_str = ln.dev.datasets.anndata_file_pbmc68k_test().as_posix()
    fp_path = Path(fp_str)

    up_str = "s3://lamindb-ci/test-data/test.csv"
    up_upath = UPath(up_str)

    _, filepath, _, _, _ = process_data(
        "id", fp_str, None, None, skip_existence_check=True
    )
    assert isinstance(filepath, LocalPathClasses)
    _, filepath, _, _, _ = process_data(
        "id", fp_path, None, None, skip_existence_check=True
    )
    assert isinstance(filepath, LocalPathClasses)

    _, filepath, _, _, _ = process_data(
        "id", up_str, None, None, skip_existence_check=True
    )
    assert isinstance(filepath, CloudPath)
    _, filepath, _, _, _ = process_data(
        "id", up_upath, None, None, skip_existence_check=True
    )
    assert isinstance(filepath, CloudPath)


def test_load_to_memory():
    # tsv
    pd.DataFrame([1, 2]).to_csv("test.tsv", sep="\t")
    df = read_tsv("test.tsv")
    assert isinstance(df, pd.DataFrame)
    # fcs
    adata = read_fcs(ln.dev.datasets.file_fcs())
    assert isinstance(adata, ad.AnnData)
    # other
    pd.DataFrame([1, 2]).to_csv("test.zrad", sep="\t")
    with pytest.raises(NotADirectoryError):
        load_to_memory("test.zrad")
    # none
    pd.DataFrame([1, 2]).to_csv("test.zip", sep="\t")
    load_to_memory("test.zip")
    assert get_hash("test.zip", suffix=".zip", check_hash=False)[0] is None
    UPath("test.tsv").unlink()
    UPath("test.zrad").unlink()
    UPath("test.zip").unlink()

    with pytest.raises(NotImplementedError) as error:
        ln.File(True)
    assert (
        error.exconly()
        == "NotImplementedError: Do not know how to create a file object from True,"
        " pass a filepath instead!"
    )


def test_delete_storage():
    with pytest.raises(FileNotFoundError):
        delete_storage(UPath("test"))


def test_describe():
    ln.dev.datasets.file_mini_csv()
    file = ln.File("mini.csv", description="test")
    file.describe()


def test_file_zarr():
    with open("test.zarr", "w") as f:
        f.write("zarr")
    assert get_hash(UPath("test.zarr"), suffix=".zarr") is None
    file = ln.File("test.zarr", description="test-zarr")
    with pytest.raises(RuntimeError) as error:
        file.stage()
    assert (
        error.exconly()
        == "RuntimeError: zarr object can't be staged, please use load() or stream()"
    )  # noqa
    file.save()
    file.delete(storage=False)
    UPath("test.zarr").unlink()


def test_zarr_folder_upload():
    previous_storage = ln.setup.settings.storage.root_as_str
    ln.settings.storage = "s3://lamindb-test"

    def callback(*args, **kwargs):
        pass

    zarr_path = Path("./test_adata.zrad")
    write_adata_zarr(adata, zarr_path, callback)

    file = ln.File(zarr_path, key="test_adata.zrad")
    file.save()

    assert isinstance(file.path, CloudPath) and file.path.exists()

    file.delete(storage=True)
    delete_storage(zarr_path)
    ln.settings.storage = previous_storage


def test_df_suffix():
    file = ln.File(df, key="test_.parquet")
    assert file.suffix == ".parquet"

    with pytest.raises(ValueError) as error:
        file = ln.File(df, key="test_.def")
    assert (
        error.exconly().partition(",")[0]
        == "ValueError: The suffix '.def' of the provided key is incorrect"
    )


def test_adata_suffix():
    file = ln.File(adata, key="test_.h5ad")
    assert file.suffix == ".h5ad"
    file = ln.File(adata, format="h5ad", key="test_.h5ad")
    assert file.suffix == ".h5ad"
    file = ln.File(adata, key="test_.zarr")
    assert file.suffix == ".zarr"
    file = ln.File(adata, key="test_.zrad")
    assert file.suffix == ".zrad"
    file = ln.File(adata, format="zrad", key="test_.zrad")
    assert file.suffix == ".zrad"

    with pytest.raises(ValueError) as error:
        file = ln.File(adata, key="test_.def")
    assert (
        error.exconly().partition(",")[0]
        == "ValueError: Error when specifying AnnData storage format"
    )

    with pytest.raises(ValueError) as error:
        file = ln.File(adata, format="h5ad", key="test.zrad")
    assert (
        error.exconly().partition(",")[0]
        == "ValueError: The suffix '.zrad' of the provided key is incorrect"
    )

    with pytest.raises(ValueError) as error:
        file = ln.File(adata, key="test_")
    assert (
        error.exconly().partition(",")[0]
        == "ValueError: The suffix '' of the provided key is incorrect"
    )



"""Storage tools.

.. autosummary::
   :toctree: .

   AnnDataAccessor
   BackedAccessor
   UPath

"""
from lamindb_setup.dev.upath import LocalPathClasses, UPath, infer_filesystem

from ._anndata_sizes import size_adata
from ._backed_access import AnnDataAccessor, BackedAccessor
from .file import delete_storage, load_to_memory, store_object
from .object import infer_suffix, write_to_file



from pathlib import Path
from typing import Optional, Union
from uuid import UUID

from lamin_utils import logger

from lamindb_setup.dev.upath import UPath
from ._close import close as close_instance
from ._init_instance import load_from_isettings
from ._settings import InstanceSettings, settings
from ._silence_loggers import silence_loggers
from .dev._settings_load import load_instance_settings
from .dev._settings_storage import StorageSettings
from .dev._settings_store import instance_settings_file
from .dev.cloud_sqlite_locker import (
    unlock_cloud_sqlite_upon_exception,
)


# this is for testing purposes only
# set to True only to test failed load
_TEST_FAILED_LOAD = False


@unlock_cloud_sqlite_upon_exception(ignore_prev_locker=True)
def load(
    identifier: str,
    *,
    storage: Optional[Union[str, Path, UPath]] = None,
    _raise_not_reachable_error: bool = True,
    _test: bool = False,
) -> Optional[str]:
    """Load existing instance.

    Args:
        identifier: `str` - The instance identifier `owner/name`.
            You can also pass the URL: `https://lamin.ai/owner/name`.
            If the instance is owned by you,
            it suffices to pass the instance name.
        storage: `Optional[PathLike] = None` - Load the instance with an
            updated default storage.
    """
    from ._check_instance_setup import check_instance_setup
    from .dev._hub_core import load_instance as load_instance_from_hub

    owner, name = get_owner_name_from_identifier(identifier)

    if check_instance_setup() and not _test:
        raise RuntimeError(
            "Currently don't support init or load of multiple instances in the same"
            " Python session. We will bring this feature back at some point."
        )
    else:
        # compare normalized identifier with a potentially previously loaded identifier
        if (
            settings._instance_exists
            and f"{owner}/{name}" != settings.instance.identifier
        ):
            close_instance(mute=True)

    # the following will return a string if the instance does not exist
    # on the hub
    hub_result = load_instance_from_hub(owner=owner, name=name)

    # if hub_result is not a string, it means it made a request
    # that successfully returned metadata
    if not isinstance(hub_result, str):
        instance_result, storage_result = hub_result
        isettings = InstanceSettings(
            owner=owner,
            name=name,
            storage_root=storage_result["root"],
            storage_region=storage_result["region"],
            db=instance_result["db"],
            schema=instance_result["schema_str"],
            id=UUID(instance_result["id"]),
        )
    else:
        settings_file = instance_settings_file(name, owner)
        if settings_file.exists():
            isettings = load_instance_settings(settings_file)
            if isettings.is_remote:
                if _raise_not_reachable_error:
                    raise RuntimeError(
                        f"Remote instance {owner}/{name} not loadable from hub. The"
                        " instance might have been deleted or you may have lost"
                        " access."
                    )
                return "instance-not-reachable"
            logger.info(f"found cached instance metadata: {settings_file}")
        else:
            if _raise_not_reachable_error:
                raise RuntimeError(
                    f"Instance {owner}/{name} neither loadable from hub nor local"
                    " cache. check whether instance exists and you have access:"
                    f" https://lamin.ai/{owner}/{name}?tab=collaborators"
                )
            return "instance-not-reachable"

    if storage is not None:
        update_isettings_with_storage(isettings, storage)
    if _test:
        isettings._persist()  # this is to test the settings
        return None
    silence_loggers()
    check, msg = isettings._load_db(
        do_not_lock_for_laminapp_admin=True
    )  # this also updates local SQLite
    if not check:
        local_db = isettings._is_cloud_sqlite and isettings._sqlite_file_local.exists()
        if local_db:
            logger.warning(
                "SQLite file does not exist in the cloud, but exists locally:"
                f" {isettings._sqlite_file_local}\nTo push the file to the cloud, call:"
                " lamin close"
            )
        elif _raise_not_reachable_error:
            raise RuntimeError(msg)
        else:
            logger.warning(
                "Instance metadata exists, but DB might have been corrupted or deleted."
                " Re-initializing the DB."
            )
            return "instance-not-reachable"
    # this is for testing purposes only
    if _TEST_FAILED_LOAD:
        raise RuntimeError("Technical testing error.")

    if storage is not None and isettings.dialect == "sqlite":
        update_root_field_in_default_storage(isettings)
    load_from_isettings(isettings)
    return None


def get_owner_name_from_identifier(identifier: str):
    if "/" in identifier:
        if identifier.startswith("https://lamin.ai/"):
            identifier = identifier.replace("https://lamin.ai/", "")
        split = identifier.split("/")
        if len(split) > 2:
            raise ValueError(
                "The instance identifier needs to be 'owner/name', the instance name"
                " (owner is current user) or the URL: https://lamin.ai/owner/name."
            )
        owner, name = split
    else:
        owner = settings.user.handle
        name = identifier
    return owner, name


def update_isettings_with_storage(
    isettings: InstanceSettings, storage: Union[str, Path, UPath]
) -> None:
    ssettings = StorageSettings(storage)
    if ssettings.is_cloud:
        try:  # triggering ssettings.id makes a lookup in the storage table
            logger.success(f"loaded storage: {ssettings.id} / {ssettings.root_as_str}")
        except RuntimeError:
            raise RuntimeError(
                "storage not registered!\n"
                "load instance without the `storage` arg and register storage root: "
                f"`lamin set storage --storage {storage}`"
            )
    else:
        # local storage
        # assumption is you want to merely update the storage location
        isettings._storage = ssettings  # need this here already
    # update isettings in place
    isettings._storage = ssettings


# this is different from register!
# register registers a new storage location
# update_root_field_in_default_storage updates the root
# field in the default storage locations
def update_root_field_in_default_storage(isettings: InstanceSettings):
    from lnschema_core.models import Storage

    storages = Storage.objects.all()
    if len(storages) != 1:
        raise RuntimeError(
            "You have several storage locations: Can't identify in which storage"
            " location the root column is to be updated!"
        )
    storage = storages[0]
    storage.root = isettings.storage.root_as_str
    storage.save()
    logger.save(f"updated storage root {storage.id} to {isettings.storage.root_as_str}")



# Generated by Django 4.2.1 on 2023-09-15 08:26

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("lnschema_core", "0018_rename_datasetlabel_datasetulabel_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="dataset",
            name="reference",
            field=models.CharField(db_index=True, default=None, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name="dataset",
            name="reference_type",
            field=models.CharField(db_index=True, default=None, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name="transform",
            name="reference_type",
            field=models.CharField(db_index=True, default=None, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name="ulabel",
            name="reference",
            field=models.CharField(db_index=True, default=None, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name="ulabel",
            name="reference_type",
            field=models.CharField(db_index=True, default=None, max_length=255, null=True),
        ),
    ]



from collections import defaultdict
from typing import Any, Dict, Iterable, List, Optional, Union

from lamin_utils import colors, logger
from lamindb_setup.dev._docs import doc_args
from lnschema_core.models import (
    Data,
    Dataset,
    Feature,
    FeatureSet,
    File,
    Registry,
    Run,
    ULabel,
    __repr__,
    format_field_value,
)
from lnschema_core.types import StrField

from lamindb.dev._settings import settings

from .._parents import view_flow
from .._query_set import QuerySet
from ._feature_manager import (
    FeatureManager,
    dict_schema_name_to_model_name,
    get_feature_set_links,
    get_host_id_field,
    get_label_links,
    print_features,
)
from ._label_manager import LabelManager, print_labels
from ._priors import priors
from ._run_context import run_context
from .exceptions import ValidationError


def get_run(run: Optional[Run]) -> Optional[Run]:
    if run is None:
        run = run_context.run
        if run is None:
            logger.warning(
                "no run & transform get linked, consider passing a `run` or calling"
                " ln.track()"
            )
    return run


def add_transform_to_kwargs(kwargs: Dict[str, Any], run: Run):
    if run is not None:
        kwargs["transform"] = run.transform


def save_transform_run_feature_sets(self: Union[File, Dataset]) -> None:
    if self.transform is not None:
        self.transform.save()
    if self.run is not None:
        self.run.save()
    if hasattr(self, "_feature_sets"):
        saved_feature_sets = {}
        for key, feature_set in self._feature_sets.items():
            if isinstance(feature_set, FeatureSet) and feature_set._state.adding:
                feature_set.save()
                saved_feature_sets[key] = feature_set
        if len(saved_feature_sets) > 0:
            s = "s" if len(saved_feature_sets) > 1 else ""
            display_feature_set_keys = ",".join(
                f"'{key}'" for key in saved_feature_sets.keys()
            )
            logger.save(
                f"saved {len(saved_feature_sets)} feature set{s} for slot{s}:"
                f" {display_feature_set_keys}"
            )


def save_feature_set_links(self: Union[File, Dataset]) -> None:
    from lamindb._save import bulk_create

    Data = self.__class__
    if hasattr(self, "_feature_sets"):
        links = []
        host_id_field = get_host_id_field(self)
        for slot, feature_set in self._feature_sets.items():
            if isinstance(feature_set, str):
                assert len(feature_set) == 20
                feature_set_id = feature_set
            else:
                feature_set_id = feature_set.id
            kwargs = {
                host_id_field: self.id,
                "feature_set_id": feature_set_id,
                "slot": slot,
            }
            links.append(Data.feature_sets.through(**kwargs))
        bulk_create(links)


@doc_args(Data.describe.__doc__)
def describe(self: Data):
    """{}"""
    model_name = self.__class__.__name__
    msg = ""

    fields = self._meta.fields
    direct_fields = []
    foreign_key_fields = []
    for f in fields:
        if f.is_relation:
            foreign_key_fields.append(f.name)
        else:
            direct_fields.append(f.name)

    # Display Provenance
    # display line by line the foreign key fields
    from .._parents import _transform_emoji

    emojis = {
        "storage": "🗃️",
        "created_by": "👤",
        "transform": _transform_emoji(self.transform),
        "run": "👣",
        "initial_version": "🔖",
        "file": "📄",
    }
    if len(foreign_key_fields) > 0:  # always True for File and Dataset
        record_msg = f"{colors.green(model_name)}{__repr__(self, include_foreign_keys=False).lstrip(model_name)}"  # noqa
        msg += f"{record_msg}\n\n"

        msg += f"{colors.green('Provenance')}:\n  "
        related_msg = "".join(
            [
                f"{emojis.get(i, '📎')} {i}: {self.__getattribute__(i)}\n  "
                for i in foreign_key_fields
                if self.__getattribute__(i) is not None
            ]
        )
        msg += related_msg
    # input of
    if self.input_of.exists():
        values = [format_field_value(i.run_at) for i in self.input_of.all()]
        msg += f"⬇️ input_of ({colors.italic('core.Run')}): {values}\n    "
    msg = msg.rstrip("    ")
    msg += print_features(self)
    msg += print_labels(self)

    logger.print(msg)


def validate_feature(feature: Feature, records: List[Registry]) -> None:
    if not isinstance(feature, Feature):
        raise TypeError("feature has to be of type Feature")
    if feature._state.adding:
        registries = set(
            [record.__class__.__get_name_with_schema__() for record in records]
        )
        registries_str = "|".join(registries)
        msg = (
            f"ln.Feature(name='{feature.name}', type='category',"
            f" registries='{registries_str}').save()"
        )
        raise ValidationError(f"Feature not validated. If it looks correct: {msg}")


def get_labels(
    self,
    feature: Feature,
    mute: bool = False,
    flat_names: bool = False,
) -> Union[QuerySet, Dict[str, QuerySet], List]:
    """{}"""
    if not isinstance(feature, Feature):
        raise TypeError("feature has to be of type Feature")
    if feature.registries is None:
        raise ValueError("feature does not have linked labels")
    registries_to_check = feature.registries.split("|")
    if len(registries_to_check) > 1 and not mute:
        logger.warning("labels come from multiple registries!")
    qs_by_registry = {}
    for registry in registries_to_check:
        # currently need to distinguish between Label and non-Label, because
        # we only have the feature information for Label
        if registry == "core.ULabel":
            links_to_labels = get_label_links(self, registry, feature)
            label_ids = [link.ulabel_id for link in links_to_labels]
            qs_by_registry[registry] = ULabel.objects.filter(id__in=label_ids)
        else:
            qs_by_registry[registry] = getattr(
                self, self.features._accessor_by_orm[registry]
            ).all()
    if flat_names:
        # returns a flat list of names
        from .._registry import get_default_str_field

        values = []
        for v in qs_by_registry.values():
            values += v.list(get_default_str_field(v))
        return values
    if len(registries_to_check) == 1:
        return qs_by_registry[registry]
    else:
        return qs_by_registry


def add_labels(
    self,
    records: Union[Registry, List[Registry], QuerySet, Iterable],
    feature: Feature,
    *,
    field: Optional[StrField] = None,
) -> None:
    """{}"""
    if isinstance(records, (QuerySet, QuerySet.__base__)):  # need to have both
        records = records.list()
    if isinstance(records, (str, Registry)):
        records = [records]
    if not isinstance(records, List):  # avoids warning for pd Series
        records = list(records)
    if isinstance(records[0], str):  # type: ignore
        records_validated = []
        if feature.registries is not None:
            orm_dict = dict_schema_name_to_model_name(File)
            for reg in feature.registries.split("|"):
                orm = orm_dict.get(reg)
                records_validated += orm.from_values(records, field=field)

        # feature doesn't have registries and therefore can't create records from values
        # ask users to pass records
        if len(records_validated) == 0:
            raise ValueError(
                "Please pass a record (a `Registry` object), not a string, e.g., via:"
                " label"
                f" = ln.ULabel(name='{records[0]}')"  # type: ignore
            )
        records = records_validated

    if self._state.adding:
        raise ValueError("Please save the file/dataset before adding a label!")
    for record in records:
        if record._state.adding:
            raise ValidationError(
                f"{record} not validated. If it looks correct: record.save()"
            )
    validate_feature(feature, records)  # type:ignore
    records_by_registry = defaultdict(list)
    for record in records:
        records_by_registry[record.__class__.__get_name_with_schema__()].append(record)
    for registry_name, records in records_by_registry.items():
        getattr(self, self.features._accessor_by_orm[registry_name]).add(
            *records, through_defaults={"feature_id": feature.id}
        )
    feature_set_links = get_feature_set_links(self)
    feature_set_ids = [link.feature_set_id for link in feature_set_links.all()]
    # get all linked features of type Feature
    feature_sets = FeatureSet.filter(id__in=feature_set_ids).all()
    linked_features_by_slot = {
        feature_set_links.filter(feature_set_id=feature_set.id)
        .one()
        .slot: feature_set.features.all()
        for feature_set in feature_sets
        if "core.Feature" == feature_set.registry
    }
    for registry_name, records in records_by_registry.items():
        msg = ""
        if feature.registries is None or registry_name not in feature.registries:
            if len(msg) > 0:
                msg += ", "
            msg += f"linked feature '{feature.name}' to registry '{registry_name}'"
            if feature.registries is None:
                feature.registries = registry_name
            elif registry_name not in feature.registries:
                feature.registries += f"|{registry_name}"
            feature.save()
        if len(msg) > 0:
            logger.save(msg)
        # check whether we have to update the feature set that manages labels
        # (Feature) to account for a new feature
        found_feature = False
        for _, linked_features in linked_features_by_slot.items():
            if feature in linked_features:
                found_feature = True
        if not found_feature:
            if "external" in linked_features_by_slot:
                feature_set = self.features._feature_set_by_slot["external"]
                features_list = feature_set.features.list()
            else:
                features_list = []
            features_list.append(feature)
            feature_set = FeatureSet(features_list, modality=priors.modalities.meta)
            feature_set.save()
            if "external" in linked_features_by_slot:
                old_feature_set_link = feature_set_links.filter(slot="external").one()
                old_feature_set_link.delete()
                remaining_links = self.feature_sets.through.objects.filter(
                    feature_set_id=feature_set.id
                ).all()
                if len(remaining_links) == 0:
                    old_feature_set = FeatureSet.filter(
                        id=old_feature_set_link.feature_set_id
                    ).one()
                    logger.info(
                        "no file links to it anymore, deleting feature set"
                        f" {old_feature_set}"
                    )
                    old_feature_set.delete()
            self.features.add_feature_set(feature_set, slot="external")
            logger.save(
                f"linked new feature '{feature.name}' together with new feature set"
                f" {feature_set}"
            )


def _track_run_input(
    data: Data, is_run_input: Optional[bool] = None, run: Optional[Run] = None
):
    if run is None:
        run = run_context.run
    track_run_input = False
    if is_run_input is None:
        # we need a global run context for this to work
        if run is not None:
            # avoid cycles (a file is both input and output)
            if data.run != run:
                if settings.track_run_inputs:
                    transform_note = ""
                    if data.transform is not None:
                        transform_note = (
                            f", adding parent transform {data.transform.id}"
                        )
                    logger.info(
                        f"adding file {data.id} as input for run"
                        f" {run.id}{transform_note}"
                    )
                    track_run_input = True
                else:
                    logger.hint(
                        "track this file as a run input by passing `is_run_input=True`"
                    )
        else:
            if settings.track_run_inputs:
                logger.hint(
                    "you can auto-track this file as a run input by calling"
                    " `ln.track()`"
                )
    else:
        track_run_input = is_run_input
    if track_run_input:
        if run is None:
            raise ValueError(
                "No run context set. Call ln.track() or link input to a"
                " run object via `run.input_files.add(file)`"
            )
        # avoid adding the same run twice
        # avoid cycles (a file is both input and output)
        if not data.input_of.contains(run) and data.run != run:
            run.save()
            data.input_of.add(run)
            if data.transform is not None:
                run.transform.parents.add(data.transform)


@property  # type: ignore
@doc_args(Data.features.__doc__)
def features(self) -> "FeatureManager":
    """{}"""
    from lamindb.dev._feature_manager import FeatureManager

    return FeatureManager(self)


@property  # type: ignore
@doc_args(Data.labels.__doc__)
def labels(self) -> "LabelManager":
    """{}"""
    from lamindb.dev._label_manager import LabelManager

    return LabelManager(self)


setattr(Data, "features", features)
setattr(Data, "labels", labels)
setattr(Data, "describe", describe)
setattr(Data, "view_flow", view_flow)



import os
from pathlib import Path
from subprocess import DEVNULL, getoutput, run


# this function is duplicated across laminhub-rest and lamindb-setup
#
def create_and_set_local_supabase_env():
    start_supabase = """supabase start -x realtime,storage-api,imgproxy,pgadmin-schema-diff,migra,postgres-meta,studio,edge-runtime,logflare,vector,pgbouncer"""  # noqa
    # unfortunately, supabase status -o env does not work with
    # a reduced number of containers (running supabase CLI version 1.38.6 & 1.96.4)
    # hence, we need this ugly regex
    get_anon_key = """grep 'anon key'|cut -f2 -d ":" | sed -e 's/^[[:space:]]*//'"""
    anon_key = getoutput(f"{start_supabase}|{get_anon_key}").split("\n")[-1]
    env = {
        "LAMIN_ENV": "local",
        "POSTGRES_DSN": "postgresql://postgres:postgres@localhost:54322/postgres",
        "SUPABASE_API_URL": "http://localhost:54321",
        "SUPABASE_ANON_KEY": anon_key,
    }
    # update environment variables with these values
    for key, value in env.items():
        # the following will not overwrite existing environment variables
        # the reason is that create_and_set_local_supabase_env seems to be called
        # multiple times; for any but the 1st time the supabase CLI is called,
        # we see a trivial output message and cannot parse the anon_key
        # (Alex doesn't understand why it's called several times and extensive
        #  debugging with logging didn't yield a conclusion)
        if key not in os.environ:
            os.environ[key] = value
        else:
            print(f"WARNING: env variable {key} is already set to {os.environ[key]}")


def pytest_configure():
    create_and_set_local_supabase_env()
    import laminhub_rest

    run(
        "lnhub alembic upgrade head",
        shell=True,
        env=os.environ,
        cwd=Path(laminhub_rest.__file__).parent.parent,
        check=True,
    )


def pytest_unconfigure():
    print("tear down supabase")
    run("supabase stop", shell=True, stdout=DEVNULL)



from pathlib import Path
from typing import Any, Optional, Union

from appdirs import AppDirs

from .upath import LocalPathClasses, UPath, create_path

DIRS = AppDirs("lamindb", "laminlabs")


class StorageSettings:
    """Manage cloud or local storage settings."""

    def __init__(
        self,
        root: Union[str, Path, UPath],
        region: Optional[str] = None,
    ):
        root_path = create_path(root)
        # root_path is either Path or UPath at this point
        if isinstance(root_path, LocalPathClasses):  # local paths
            # resolve fails for nonexisting dir
            root_path.mkdir(parents=True, exist_ok=True)
            root_path = root_path.resolve()
        self._root = root_path
        self._region = region
        # would prefer to type below as Registry, but need to think through import order
        self._record: Optional[Any] = None

    @property
    def id(self) -> str:
        """Storage id."""
        return self.record.id

    @property
    def record(self) -> Any:
        """Storage record."""
        if self._record is None:
            # dynamic import because of import order
            from lnschema_core.models import Storage

            self._record = Storage.objects.get(root=self.root_as_str)
        return self._record

    @property
    def root(self) -> UPath:
        """Root storage location."""
        return self._root

    def _set_fs_kwargs(self, **kwargs):
        """Set additional fsspec arguments for cloud root.

        Example:

        >>> ln.setup.settings.storage._set_fs_kwargs(  # any fsspec args
        >>>    profile="some_profile", cache_regions=True
        >>> )
        """
        if not isinstance(self._root, LocalPathClasses):
            self._root = UPath(self._root, **kwargs)

    @property
    def root_as_str(self) -> str:
        """Formatted root string."""
        return self.root.as_posix().rstrip("/")

    @property
    def cache_dir(
        self,
    ) -> Path:
        """Cache root, a local directory to cache cloud files."""
        cache_dir = UPath(DIRS.user_cache_dir)
        cache_dir.mkdir(parents=True, exist_ok=True)
        return cache_dir

    @property
    def is_cloud(self) -> bool:
        """`True` if `storage_root` is in cloud, `False` otherwise."""
        return not isinstance(self.root, LocalPathClasses)

    @property
    def region(self) -> Optional[str]:
        """Storage region."""
        return self._region

    @property
    def type(self) -> str:
        """AWS S3 vs. Google Cloud vs. local vs. the other protocols.

        Returns the protocol.
        """
        import fsspec

        convert = {"file": "local"}
        protocol = fsspec.utils.get_protocol(str(self.root))
        return convert.get(protocol, protocol)

    def key_to_filepath(self, filekey: Union[Path, UPath, str]) -> UPath:
        """Cloud or local filepath from filekey."""
        return self.root / filekey

    def cloud_to_local(
        self, filepath: Union[Path, UPath], **kwargs
    ) -> Union[Path, UPath]:
        """Local (cache) filepath from filepath."""
        local_filepath = self.cloud_to_local_no_update(filepath)  # type: ignore
        if isinstance(filepath, UPath) and not isinstance(filepath, LocalPathClasses):
            local_filepath.parent.mkdir(parents=True, exist_ok=True)
            filepath.synchronize(local_filepath, **kwargs)
        return local_filepath

    # conversion to Path via cloud_to_local() would trigger download
    # of remote file to cache if there already is one
    # in pure write operations that update the cloud, we don't want this
    # hence, we manually construct the local file path
    # using the `.parts` attribute in the following line
    def cloud_to_local_no_update(
        self, filepath: Union[Path, UPath]
    ) -> Union[Path, UPath]:
        if isinstance(filepath, UPath) and not isinstance(filepath, LocalPathClasses):
            return self.cache_dir.joinpath(filepath._url.netloc, *filepath.parts[1:])  # type: ignore # noqa
        return filepath

    def local_filepath(self, filekey: Union[Path, UPath, str]) -> UPath:
        """Local (cache) filepath from filekey: `local(filepath(...))`."""
        return self.cloud_to_local(self.key_to_filepath(filekey))


# below is for backward compat
Storage = StorageSettings



import shutil

import h5py
import pandas as pd
import pytest
import zarr

import lamindb as ln
from lamindb.dev.storage import delete_storage
from lamindb.dev.storage._backed_access import backed_access
from lamindb.dev.storage._zarr import read_adata_zarr, write_adata_zarr
from lamindb.dev.storage.file import read_adata_h5ad
from lamindb.dev.storage.object import infer_suffix, write_to_file


@pytest.fixture
def bad_adata_path():
    fp = ln.dev.datasets.anndata_file_pbmc68k_test()
    adata = read_adata_h5ad(fp)
    to = fp.with_name("pbmc68k_bad.h5ad")
    shutil.copy(fp, to)
    fp = to
    file = h5py.File(fp, mode="r+")
    for field_name in ("obs", "var"):
        field = getattr(adata, field_name).to_records()
        formats = []
        for name, (dt, _) in field.dtype.fields.items():
            if dt == "O":
                new_dt = str(field[name].astype(str).dtype).replace("<U", "S")
            else:
                new_dt = dt
            formats.append((name, new_dt))
        del file[field_name]
        file.create_dataset(field_name, data=field.astype(formats))
    del file["X"].attrs["encoding-type"]
    del file["X"].attrs["encoding-version"]
    del file["obsp"]["test"].attrs["encoding-type"]
    del file["obsp"]["test"].attrs["encoding-version"]
    file.close()
    return fp


def test_anndata_io():
    test_file = ln.dev.datasets.anndata_file_pbmc68k_test()

    adata = read_adata_h5ad(test_file)

    def callback(*args, **kwargs):
        pass

    zarr_path = test_file.with_suffix(".zarr")
    write_adata_zarr(adata, zarr_path, callback)

    adata = read_adata_zarr(zarr_path)

    assert adata.shape == (30, 200)

    delete_storage(zarr_path)


@pytest.mark.parametrize("adata_format", ["h5ad", "zarr"])
def test_backed_access(adata_format):
    fp = ln.dev.datasets.anndata_file_pbmc68k_test()
    if adata_format == "zarr":
        adata = read_adata_h5ad(fp)

        def callback(*args, **kwargs):
            pass

        fp = fp.with_suffix(".zarr")
        write_adata_zarr(adata, fp, callback)
        del adata
        # remove encoding information to check correctness of backed accessor
        store = zarr.open(fp)
        del store["obsp"]["test"].attrs["encoding-type"]
        del store["obsp"]["test"].attrs["encoding-version"]
        del store["obsm"]["X_pca"].attrs["encoding-type"]
        del store["obsm"]["X_pca"].attrs["encoding-version"]
        del store

    with pytest.raises(ValueError):
        access = backed_access(fp.with_suffix(".invalid_suffix"))

    access = backed_access(fp)

    assert isinstance(access.obs_names, pd.Index)
    assert isinstance(access.var_names, pd.Index)
    assert access.raw.shape == (30, 100)
    assert access.obsp["test"].to_memory().sum() == 30
    assert access.varp["test"].to_memory().sum() == 200
    assert access.layers["test"][0].sum() == 200

    sub = access[:10]
    assert sub[:5].shape == (5, 200)
    assert sub.layers["test"].shape == sub.shape
    assert sub.raw.shape == (10, 100)
    assert sub.obsp["test"].sum() == 10
    assert sub.varp["test"].sum() == 200
    assert sub.obsm["X_pca"].shape == (10, 50)

    with pytest.raises(AttributeError):
        sub.raw.raw

    assert access[:, [1, 2, 5]].varp["test"].sum() == 3

    obs_sub = ["TCAATCACCCTTCG-8", "CGTTATACAGTACC-8", "TGCCAAGATTGTGG-7"]
    sub = access[obs_sub]
    assert sub.obs_names.tolist() == obs_sub

    var_sub = ["SSU72", "PARK7", "RBP7"]
    sub = access[:, var_sub]
    assert sub.var_names.tolist() == var_sub

    assert access.to_memory().shape == (30, 200)
    assert sub.to_memory().shape == (30, 3)

    access.close()
    if adata_format == "zarr":
        assert fp.suffix == ".zarr"
        delete_storage(fp)


def test_infer_suffix():
    import anndata as ad

    adata = ad.AnnData()
    assert infer_suffix(adata, adata_format="h5ad") == ".h5ad"
    with pytest.raises(ValueError):
        infer_suffix(adata, adata_format="my format")
    with pytest.raises(NotImplementedError):
        infer_suffix(ln.File)


def test_write_to_file():
    with pytest.raises(NotImplementedError):
        write_to_file(ln.File, "path")


def test_backed_bad_format(bad_adata_path):
    access = backed_access(bad_adata_path)

    assert access.obsp["test"].to_memory().sum() == 30

    sub = access[:10]

    assert sub.X.shape == (10, 200)
    assert sub.obsp["test"].sum() == 10

    assert isinstance(sub.obs, pd.DataFrame)
    assert isinstance(sub.var, pd.DataFrame)
    assert isinstance(sub.obs_names, pd.Index)
    assert isinstance(sub.var_names, pd.Index)

    assert sub.to_memory().shape == (10, 200)

    access.close()
    delete_storage(bad_adata_path)



import shutil
from subprocess import DEVNULL, run

import lamindb_setup
import pytest
from lamin_utils import logger
from laminci.db import setup_local_test_postgres


def pytest_sessionstart(session: pytest.Session):
    pgurl = setup_local_test_postgres()
    lamindb_setup.init(
        storage="./default_storage",
        schema="bionty",
        name="lamindb-unit-tests",
        db=pgurl,
    )
    # we're setting this to true prior to importing lamindb!
    lamindb_setup._TESTING = True


def pytest_sessionfinish(session: pytest.Session):
    logger.set_verbosity(1)
    lamindb_setup.delete("lamindb-unit-tests", force=True)
    shutil.rmtree("./default_storage")
    # shutil.rmtree("./outside_storage")
    run("docker stop pgtest && docker rm pgtest", shell=True, stdout=DEVNULL)



import os
import shutil
from pathlib import Path
from typing import Literal, Optional, Union

import anndata as ad
import fsspec
import pandas as pd
from lamin_utils import logger
from lamindb_setup import settings
from lamindb_setup.dev import StorageSettings
from lamindb_setup.dev.upath import (
    LocalPathClasses,
    UPath,
    create_path,
    infer_filesystem,
)
from lnschema_core.models import File, Storage

try:
    from ._zarr import read_adata_zarr
except ImportError:

    def read_adata_zarr(filepath):  # type: ignore
        raise ImportError("Please install zarr: pip install zarr")


AUTO_KEY_PREFIX = ".lamindb/"


# also see https://gist.github.com/securifera/e7eed730cbe1ce43d0c29d7cd2d582f4
#    ".gz" is not listed here as it typically occurs with another suffix
KNOWN_SUFFIXES = {
    #
    # without readers
    #
    ".fasta",
    ".fastq",
    ".jpg",
    ".mtx",
    ".obo",
    ".pdf",
    ".png",
    ".tar",
    ".tiff",
    ".txt",
    ".tsv",
    ".zip",
    ".xml",
    #
    # with readers (see below)
    #
    ".h5ad",
    ".parquet",
    ".csv",
    ".fcs",
    ".xslx",
    ".zarr",
    ".zrad",
}


def extract_suffix_from_path(
    path: Union[UPath, Path], arg_name: Optional[str] = None
) -> str:
    if len(path.suffixes) <= 1:
        return path.suffix
    else:
        arg_name = "file" if arg_name is None else arg_name  # for the warning
        msg = f"{arg_name} has more than one suffix (path.suffixes), "
        # first check the 2nd-to-last suffix because it might be followed by .gz
        # or another compression-related suffix
        # Alex thought about adding logic along the lines of path.suffixes[-1]
        # in COMPRESSION_SUFFIXES to detect something like .random.gz and then
        # add ".random.gz" but concluded it's too dangerous it's safer to just
        # use ".gz" in such a case
        if path.suffixes[-2] in KNOWN_SUFFIXES:
            suffix = "".join(path.suffixes[-2:])
            msg += f"inferring: '{suffix}'"
        else:
            suffix = path.suffixes[-1]  # this is equivalent to path.suffix!!!
            msg += (
                f"using only last suffix: '{suffix}' - if you want your file format to"
                " be recognized, make an issue:"
                " https://github.com/laminlabs/lamindb/issues/new"
            )
        logger.warning(msg)
        return suffix


# add type annotations back asap when re-organizing the module
def auto_storage_key_from_file(file: File):
    if file.key is None:
        return auto_storage_key_from_id_suffix(file.id, file.suffix)
    else:
        return file.key


def auto_storage_key_from_id_suffix(id: str, suffix: str) -> str:
    assert isinstance(id, str)
    assert isinstance(suffix, str)
    storage_key = f"{AUTO_KEY_PREFIX}{id}{suffix}"
    return storage_key


def attempt_accessing_path(file: File, storage_key: str):
    if file.storage_id == settings.storage.id:
        path = settings.storage.key_to_filepath(storage_key)
    else:
        logger.debug("file.path is slightly slower for files outside default storage")
        storage = Storage.filter(id=file.storage_id).one()
        # find a better way than passing None to instance_settings in the future!
        storage_settings = StorageSettings(storage.root)
        path = storage_settings.key_to_filepath(storage_key)
    return path


# add type annotations back asap when re-organizing the module
def filepath_from_file(file: File):
    if hasattr(file, "_local_filepath") and file._local_filepath is not None:
        return file._local_filepath.resolve()
    storage_key = auto_storage_key_from_file(file)
    path = attempt_accessing_path(file, storage_key)
    return path


def read_adata_h5ad(filepath, **kwargs) -> ad.AnnData:
    fs, filepath = infer_filesystem(filepath)

    with fs.open(filepath, mode="rb") as file:
        adata = ad.read_h5ad(file, backed=False, **kwargs)
        return adata


def print_hook(size: int, value: int, **kwargs):
    progress = value / size
    out = (
        f"... {kwargs['action']} {Path(kwargs['filepath']).name}:"
        f" {min(progress, 1.):4.2f}"
    )
    if progress >= 1:
        out += "\n"
    if "NBPRJ_TEST_NBPATH" not in os.environ:
        print(out, end="\r")


class ProgressCallback(fsspec.callbacks.Callback):
    def __init__(self, action: Literal["uploading", "downloading"]):
        super().__init__()
        self.action = action

    def branch(self, path_1, path_2, kwargs):
        kwargs["callback"] = fsspec.callbacks.Callback(
            hooks=dict(print_hook=print_hook), filepath=path_1, action=self.action
        )

    def call(self, *args, **kwargs):
        return None


def store_object(localpath: Union[str, Path, UPath], storagekey: str) -> float:
    """Store arbitrary file to configured storage location.

    Returns size in bytes.
    """
    storagepath: UPath = settings.instance.storage.key_to_filepath(storagekey)
    localpath = Path(localpath)

    if localpath.is_file():
        size = localpath.stat().st_size
    else:
        size = sum(f.stat().st_size for f in localpath.rglob("*") if f.is_file())

    if not isinstance(storagepath, LocalPathClasses):
        if localpath.suffix not in {".zarr", ".zrad"}:
            cb = ProgressCallback("uploading")
        else:
            # todo: make proper progress bar for zarr
            cb = fsspec.callbacks.NoOpCallback()
        storagepath.upload_from(localpath, recursive=True, callback=cb)
    else:  # storage path is local
        storagepath.parent.mkdir(parents=True, exist_ok=True)
        if localpath.is_file():
            try:
                shutil.copyfile(localpath, storagepath)
            except shutil.SameFileError:
                pass
        else:
            if storagepath.exists():
                shutil.rmtree(storagepath)
            shutil.copytree(localpath, storagepath)
    return float(size)  # because this is how we store in the db


def delete_storage_using_key(file: File, storage_key: str):
    filepath = attempt_accessing_path(file, storage_key)
    delete_storage(filepath)


def delete_storage(storagepath: Union[Path, UPath]):
    """Delete arbitrary file."""
    if storagepath.is_file():
        storagepath.unlink()
    elif storagepath.is_dir():
        if isinstance(storagepath, LocalPathClasses) or not isinstance(
            storagepath, UPath
        ):
            shutil.rmtree(storagepath)
        else:
            storagepath.rmdir()
    else:
        raise FileNotFoundError(f"{storagepath} is not an existing path!")


# tested in lamin-usecases
def read_fcs(*args, **kwargs):
    try:
        import readfcs
    except ImportError:  # pragma: no cover
        raise ImportError("Please install readfcs: pip install readfcs")
    return readfcs.read(*args, **kwargs)


def read_tsv(path: Union[str, Path, UPath]) -> pd.DataFrame:
    path_sanitized = Path(path)
    return pd.read_csv(path_sanitized, sep="\t")


def load_to_memory(filepath: Union[str, Path, UPath], stream: bool = False, **kwargs):
    """Load a file into memory.

    Returns the filepath if no in-memory form is found.
    """
    filepath = create_path(filepath)

    if filepath.suffix in (".zarr", ".zrad"):
        stream = True
    elif filepath.suffix != ".h5ad":
        stream = False

    if not stream:
        # caching happens here if filename is a UPath
        # todo: make it safe when filepath is just Path
        cb = ProgressCallback("downloading")
        filepath = settings.instance.storage.cloud_to_local(filepath, callback=cb)

    READER_FUNCS = {
        ".csv": pd.read_csv,
        ".tsv": read_tsv,
        ".h5ad": read_adata_h5ad,
        ".parquet": pd.read_parquet,
        ".fcs": read_fcs,
        ".zarr": read_adata_zarr,
        ".zrad": read_adata_zarr,
    }

    reader = READER_FUNCS.get(filepath.suffix)
    if reader is None:
        return filepath
    else:
        return reader(filepath, **kwargs)



import logging


# from https://github.com/boto/boto3/blob/8c6e641bed8130a9d8cb4d97b4acbe7aa0d0657a/boto3/__init__.py#L37  # noqa
def set_stream_logger(name, level):
    logger = logging.getLogger(name)
    logger.setLevel(level)
    handler = logging.StreamHandler()
    handler.setLevel(level)
    logger.addHandler(handler)


def silence_loggers():
    # this gets logged 6 times
    set_stream_logger(name="botocore.credentials", level=logging.WARNING)
    set_stream_logger(name="botocore.hooks", level=logging.WARNING)
    set_stream_logger(name="botocore.utils", level=logging.WARNING)
    set_stream_logger(name="botocore.auth", level=logging.WARNING)
    set_stream_logger(name="botocore.endpoint", level=logging.WARNING)
    try:
        import aibotocore  # noqa

        # the 7th logging message of credentials came from aiobotocore
        set_stream_logger(name="aiobotocore.credentials", level=logging.WARNING)
    except ImportError:
        pass
